import troposphere.ec2 as ec2

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


class SecurityGroup(ec2.SecurityGroup):
    # troposphere ec2 is quite old and SecurityGroupIngress is only a list
    # without the obj type, this break auto_get_props, fix it overriding
    props = {
        "GroupName": (str, False),
        "GroupDescription": (str, True),
        "SecurityGroupEgress": (list, False),
        "SecurityGroupIngress": ([ec2.SecurityGroupRule], False),
        "VpcId": (str, False),
        "Tags": ((Tags, list), False),
    }

    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.VpcId = get_expvalue("VpcId")


class SecurityGroupIngress(ec2.SecurityGroupIngress):
    IpProtocol = "tcp"


class SecurityGroupRule(ec2.SecurityGroupRule):
    IpProtocol = "tcp"


class SecurityGroupIngressInstanceELBPorts(SecurityGroupIngress):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        name = self.title  # Ex. SecurityGroupIngressListeners1
        self.GroupId = GetAtt("SecurityGroupInstancesRules", "GroupId")


class EC2Subnet(ec2.Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.AvailabilityZone = Sub("${AWS::Region}%s" % zone.lower())
        self.VpcId = Ref("VPC")


class EC2SubnetPrivate(EC2Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, zone, **kwargs)
        self.CidrBlock = Ref(f"SubnetCidrBlockPrivate{zone}")
        self.MapPublicIpOnLaunch = False
        self.Tags = Tags(Name=Sub("${VPCName}-Private%s" % zone))


class EC2SubnetPublic(EC2Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, zone, **kwargs)
        self.CidrBlock = Ref(f"SubnetCidrBlockPublic{zone}")
        self.MapPublicIpOnLaunch = True
        self.Tags = Tags(Name=Sub("${VPCName}-Public%s" % zone))


class EC2SubnetRouteTableAssociationPrivate(ec2.SubnetRouteTableAssociation):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.RouteTableId = Ref("RouteTablePrivate")
        self.SubnetId = Ref(f"SubnetPrivate{zone}")


class EC2SubnetRouteTableAssociationPublic(ec2.SubnetRouteTableAssociation):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.RouteTableId = Ref("RouteTablePublic")
        self.SubnetId = Ref(f"SubnetPublic{zone}")


def SG_SecurityGroupsExtra(Out_String, Out_Map):
    # Parameters
    P_SecurityGroups = Parameter(
        "SecurityGroups",
        Description=f"SecurityGroups List Extra - {SECURITY_GROUPS_DEFAULT} for default based on env/role",
        AllowedPattern=r"^(\w*,\w*){%s}$" % (MAX_SECURITY_GROUPS - 1),
        Default=SECURITY_GROUPS_DEFAULT,
    )

    add_obj([P_SecurityGroups])

    SecurityGroups = []

    for n in range(MAX_SECURITY_GROUPS):
        name = f"SecurityGroup{n}"  # Ex SecurityGroup1
        value = Select(n, Split(",", get_endvalue("SecurityGroups")))
        outnamename = f"SecurityGroupName{n}"
        outvaluename = f"SecurityGroupValue{n}"

        # conditions
        add_obj(
            {
                name: Not(
                    get_condition(
                        "", "equals", "none", Select(n, Split(",", "SecurityGroups"))
                    )
                )
            }
        )

        SecurityGroups.append(
            If(name, get_expvalue(value, prefix="SecurityGroup"), Ref("AWS::NoValue"))
        )

        # outputs
        Out_String.append("${%s}=${%s}" % (outnamename, outvaluename))
        Out_Map.update(
            {
                outnamename: value,
                outvaluename: If(
                    name, get_expvalue(value, prefix="SecurityGroup"), "none"
                ),
            }
        )

    # Outputs
    O_SecurityGroups = Output(
        "SecurityGroups", Value=Sub(",".join(Out_String), **Out_Map)
    )

    add_obj(O_SecurityGroups)

    cfg.SecurityGroupsImport = SecurityGroups


def SG_SecurityGroupsEC2(key):
    Out_String = ["Rules=${SecurityGroupInstancesRules}"]
    Out_Map = {"SecurityGroupInstancesRules": {"Ref": "SecurityGroupInstancesRules"}}

    SecurityGroups = SG_SecurityGroupsExtra(Out_String, Out_Map)


def SG_SecurityGroupsECS(key):
    Out_String = ["Service=${SecurityGroupEcsService}"]
    Out_Map = {"SecurityGroupEcsService": {"Ref": "SecurityGroupEcsService"}}

    SecurityGroups = SG_SecurityGroupsExtra(Out_String, Out_Map)
    # add Condition to Output created by SG_SecurityGroupsExtra
    try:
        cfg.Outputs["SecurityGroups"].Condition = "ECSTaskDefinitionBaseNetworkModeAwsVpc"
    except Exception:
        pass


def SG_SecurityGroupsTSK(key):
    Out_String = []
    Out_Map = {}

    SecurityGroups = SG_SecurityGroupsExtra(Out_String, Out_Map)
    # add Condition to Output created by SG_SecurityGroupsExtra
    try:
        cfg.Outputs["SecurityGroups"].Condition = "ECSTaskDefinitionBaseNetworkModeAwsVpc"
    except Exception:
        pass


def SG_SecurityGroupRules(groupname, ingresses):
    SecurityGroup_Rules = []
    listeners_cfg = {}

    if cfg.LoadBalancer:
        if cfg.LoadBalancerType == "Network":
            return SecurityGroup_Rules

        if cfg.LoadBalancerType == "Classic":
            # get config from ElasticLoadBalancingLoadBalancer key
            for e in cfg.LoadBalancer:
                listeners_cfg.update(
                    getattr(cfg, f"ElasticLoadBalancingLoadBalancer{e}")["Listeners"]
                )
        if cfg.LoadBalancerType == "Application":
            # get config ElasticLoadBalancingV2Listener key
            for n, v in cfg.ElasticLoadBalancingV2Listener.items():
                if v.get("IBOX_ENABLED", True):
                    listeners_cfg[n] = v

    # Trick to populate SecurityGroupIngress using Listeners
    if ingresses:
        # use SecurityGroupIngress
        prefix = f"{groupname}SecurityGroupIngress"
        use_listener = False
        cond_key = "CidrIp"
    else:
        # use cfg.Listeners
        prefix = "Listeners"
        ingresses = listeners_cfg
        use_listener = True
        cond_key = "Access"

    for n, v in ingresses.items():
        if use_listener:
            # Trick to populate SecurityGroupIngress using Listeners
            sg_rootdict = {
                "CidrIp": v.get("Access", "Public"),
                "FromPort": v["LoadBalancerPort"]
                if "LoadBalancerPort" in v
                else v["Port"],
                "ToPort": v["LoadBalancerPort"]
                if "LoadBalancerPort" in v
                else v["Port"],
            }
            kwargs = {"rootdict": sg_rootdict}
        else:
            sg_rootdict = v
            kwargs = {}

        resname = f"{prefix}{n}"
        allowed_ip = sg_rootdict.get("CidrIp") == "AllowedIp"
        allowed_ip_or_public = sg_rootdict.get("AllowedIpOrPublic")
        if allowed_ip:
            for m, w in cfg.AllowedIp.items():
                r_SGRule = SecurityGroupRule(resname)
                auto_get_props(
                    r_SGRule, **kwargs, res_obj_type="AWS::EC2::SecurityGroup"
                )
                auto_get_props(
                    r_SGRule, f"AllowedIp{m}", res_obj_type="AWS::EC2::SecurityGroup"
                )
                SecurityGroup_Rules.append(
                    If(f"AllowedIp{m}", r_SGRule, Ref("AWS::NoValue"))
                )

        if not allowed_ip or allowed_ip_or_public:
            r_SGRule = SecurityGroupRule(resname)
            auto_get_props(r_SGRule, **kwargs, res_obj_type="AWS::EC2::SecurityGroup")

            if allowed_ip and allowed_ip_or_public:
                r_SGRule.CidrIp = "0.0.0.0/0"
                # condition
                c_Public = get_condition(
                    f"{resname}Public", "equals", "0.0.0.0/0", f"{resname}{cond_key}"
                )
                add_obj(c_Public)
                r_SGRule = If(f"{resname}Public", r_SGRule, Ref("AWS::NoValue"))

            SecurityGroup_Rules.append(r_SGRule)

    return SecurityGroup_Rules


def SG_SecurityGroup(key):
    for n, v in getattr(cfg, key).items():
        if not v.get("IBOX_ENABLED", True):
            continue
        # harcode original title, as is used this way in other part of code/cfg
        mapname = f"{key}{n}"
        resname = f"SecurityGroup{n}"

        # resources
        r_SG = SecurityGroup(resname)
        auto_get_props(r_SG, mapname=mapname)

        try:
            ingress = v["SecurityGroupIngress"]
            if ingress:
                # SecurityGroupIngress is not empty, check if is "complete" (have IpProtocol)
                r_SG.to_dict()
            else:
                raise ValueError
        except KeyError:
            # key SecurityGroupIngress do not exist
            pass
        except ValueError:
            # to_dict failed or SecurityGroupIngress is empty, populate it using SG_SecurityGroupRules
            r_SG.SecurityGroupIngress = SG_SecurityGroupRules(mapname, v["SecurityGroupIngress"])

        try:
            outname = v["OutputName"]
        except Exception:
            outname = resname
        else:
            outname = f"SecurityGroup{outname}"

        # outputs
        o_SG = Output(outname)
        o_SG.Value = GetAtt(resname, "GroupId")
        if v.get("Export"):
            o_SG.Export = Export(outname)

        add_obj(r_SG)
        # add output only if not already present (can be created by IBOXOUTPUT)
        try:
            cfg.Outputs[outname]
        except Exception:
            if not v.get("IBOX_NO_OUTPUT"):
                add_obj(o_SG)


def SG_SecurityGroupIngresses(key):
    for n, v in getattr(cfg, key).items():
        if not v.get("IBOX_ENABLED", True):
            continue
        mapname = f"{key}{n}"
        resname = f"SecurityGroupIngress{n}"
        try:
            allowed_ip = v["CidrIp"] == "AllowedIp"
        except Exception:
            pass
        else:
            if allowed_ip:
                for m, w in cfg.AllowedIp.items():
                    r_SGI = SecurityGroupIngress(f"{resname}{m}")
                    auto_get_props(r_SGI, mapname)
                    auto_get_props(r_SGI, f"AllowedIp{m}")
                    r_SGI.Condition = f"AllowedIp{m}"
                    add_obj(r_SGI)
                continue

        r_SGI = SecurityGroupIngress(resname)
        auto_get_props(r_SGI, mapname)
        add_obj(r_SGI)


def EC2_Subnet(key):
    o_subnetprivate = []
    o_subnetpublic = []

    for i in range(cfg.AZones["MAX"]):
        zone_name = cfg.AZoneNames[i]
        zone_cond = f"Zone{zone_name}"

        # parameters
        p_SubnetCidrBlockPrivate = Parameter(
            f"SubnetCidrBlockPrivate{zone_name}",
            Description=f"Ip Class Range for Private Subnet in Zone {zone_name}",
            Default=cfg.VPC_DEFAULT_SUBNETS_CIDR_BLOCK_PRIVATE[i],
        )

        p_SubnetCidrBlockPublic = Parameter(
            f"SubnetCidrBlockPublic{zone_name}",
            Description=f"Ip Class Range for Public Subnet in zone {zone_name}",
            Default=cfg.VPC_DEFAULT_SUBNETS_CIDR_BLOCK_PUBLIC[i],
        )

        add_obj([p_SubnetCidrBlockPrivate, p_SubnetCidrBlockPublic])

        # conditions
        c_Zone = {
            zone_cond: Equals(
                FindInMap("AvabilityZones", Ref("AWS::Region"), f"Zone{i}"), "True"
            )
        }

        add_obj(c_Zone)

        # resources

        r_SubnetPrivate = EC2SubnetPrivate(
            f"SubnetPrivate{zone_name}", zone=zone_name, Condition=zone_cond
        )

        r_SubnetPublic = EC2SubnetPublic(
            f"SubnetPublic{zone_name}", zone=zone_name, Condition=zone_cond
        )

        r_SubnetRouteTableAssociationPrivate = EC2SubnetRouteTableAssociationPrivate(
            f"SubnetRouteTableAssociationPrivate{zone_name}",
            zone=zone_name,
            Condition=zone_cond,
        )

        r_SubnetRouteTableAssociationPublic = EC2SubnetRouteTableAssociationPublic(
            f"SubnetRouteTableAssociationPublic{zone_name}",
            zone=zone_name,
            Condition=zone_cond,
        )

        add_obj(
            [
                r_SubnetPrivate,
                r_SubnetPublic,
                r_SubnetRouteTableAssociationPrivate,
                r_SubnetRouteTableAssociationPublic,
            ]
        )

        # outputs
        o_subnetprivate.append(
            If(zone_cond, Ref(f"SubnetPrivate{zone_name}"), Ref("AWS::NoValue"))
        )

        o_subnetpublic.append(
            If(zone_cond, Ref(f"SubnetPublic{zone_name}"), Ref("AWS::NoValue"))
        )

    # Outputs
    O_SubnetsPrivate = Output(
        "SubnetsPrivate",
        Value=Join(",", o_subnetprivate),
        Export=Export("SubnetsPrivate"),
    )

    O_SubnetsPublic = Output(
        "SubnetsPublic", Value=Join(",", o_subnetpublic), Export=Export("SubnetsPublic")
    )

    add_obj([O_SubnetsPrivate, O_SubnetsPublic])



import os
import sys
import random
import string
import copy
import json
import logging
from pprint import pprint, pformat
from pathlib import PurePath
from troposphere import validators
from troposphere.autoscaling import Tags as asgTags
import troposphere.ssm as ssm

from troposphere import (
    And,
    AWSHelperFn,
    AWSObject,
    AWSProperty,
    PropsDictType,
    Base64,
    BaseAWSObject,
    Condition,
    Equals,
    Export,
    FindInMap,
    GetAtt,
    GetAZs,
    If,
    ImportValue,
    Join,
    Not,
    Or,
    Output,
    Parameter,
    Ref,
    Select,
    Split,
    Sub,
    Tags,
    Template,
)

from . import cfg

from .cfg import (
    MAX_SECURITY_GROUPS,
    SECURITY_GROUPS_DEFAULT,
)


# Temporary fix for https://github.com/cloudtools/troposphere/issues/1474
def my_one_of(class_name, properties, property, conditionals):
    if properties.get(property) not in conditionals and not isinstance(
        properties.get(property), If
    ):
        raise ValueError(
            # Ensure we handle None as a valid value
            '%s.%s must be one of: "%s"'
            % (
                class_name,
                property,
                ", ".join(condition for condition in conditionals if condition),
            )
        )


def boolean(x):
    if x in [True, 1, "1", "true", "True"]:
        return True
    if x in [False, 0, "0", "false", "False"]:
        return False
    raise ValueError


# Fix to avoid setting OS variable TROPO_REAL_BOOL
validators.boolean = boolean



import troposphere.rds as rds

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


class RDSDBInstance(rds.DBInstance):
    def validate(self):
        if "DBSnapshotIdentifier" not in self.properties:
            if "Engine" not in self.properties:
                raise ValueError(
                    "Resource Engine is required in type %s" % self.resource_type
                )

        #       if 'SourceDBInstanceIdentifier' in self.properties:
        #
        #           invalid_replica_properties = (
        #               'BackupRetentionPeriod', 'DBName', 'MasterUsername',
        #               'MasterUserPassword', 'PreferredBackupWindow', 'MultiAZ',
        #               'DBSnapshotIdentifier',
        #           )
        #
        #           invalid_properties = [s for s in self.properties.keys() if
        #                                 s in invalid_replica_properties]
        #
        #           if invalid_properties:
        #                raise ValueError(
        #                   ('{0} properties can\'t be provided when '
        #                    'SourceDBInstanceIdentifier is present '
        #                    'AWS::RDS::DBInstance.'
        #                    ).format(', '.join(sorted(invalid_properties))))

        if (
            (
                "DBSnapshotIdentifier" not in self.properties
                and "SourceDBInstanceIdentifier" not in self.properties
            )
            and (
                "MasterUsername" not in self.properties
                or "MasterUserPassword" not in self.properties
            )
            and ("DBClusterIdentifier" not in self.properties)
        ):
            raise ValueError(
                "Either (MasterUsername and MasterUserPassword) or"
                " DBSnapshotIdentifier are required in type "
                "AWS::RDS::DBInstance."
            )

        if "KmsKeyId" in self.properties and "StorageEncrypted" not in self.properties:
            raise ValueError(
                "If KmsKeyId is provided, StorageEncrypted is required "
                "AWS::RDS::DBInstance."
            )

        nonetype = type(None)
        avail_zone = self.properties.get("AvailabilityZone", None)
        multi_az = self.properties.get("MultiAZ", None)
        if not (
            isinstance(avail_zone, (AWSHelperFn, nonetype))
            and isinstance(multi_az, (AWSHelperFn, nonetype))
        ):
            if avail_zone and multi_az in [True, 1, "1", "true", "True"]:
                raise ValueError(
                    "AvailabiltyZone cannot be set on "
                    "DBInstance if MultiAZ is set to true."
                )

        storage_type = self.properties.get("StorageType", None)
        if storage_type and storage_type == "io1" and "Iops" not in self.properties:
            raise ValueError("Must specify Iops if using StorageType io1")

        allocated_storage = self.properties.get("AllocatedStorage")
        iops = self.properties.get("Iops", None)
        if iops and not isinstance(iops, AWSHelperFn):
            if (
                not isinstance(allocated_storage, AWSHelperFn)
                and allocated_storage < 100
            ):
                raise ValueError(
                    "AllocatedStorage must be at least 100 when " "Iops is set."
                )
            if (
                not isinstance(allocated_storage, AWSHelperFn)
                and not isinstance(iops, AWSHelperFn)
                and float(iops) / float(allocated_storage) > 10.0
            ):
                raise ValueError(
                    "AllocatedStorage must be no less than "
                    "1/10th the provisioned Iops"
                )

        return True


class RDSDBParameterGroup(rds.DBParameterGroup):
    def __init__(self, title, dbinstance, **kwargs):
        super().__init__(title, **kwargs)
        self.Description = Sub("MYSQL %s - ${AWS::StackName}" % self.title)
        self.Family = Sub(
            "${Engine}${EngineVersion}",
            **{
                "Engine": get_endvalue(f"{dbinstance}Engine"),
                "EngineVersion": Join(
                    ".",
                    [
                        Select(
                            0, Split(".", get_endvalue(f"{dbinstance}EngineVersion"))
                        ),
                        Select(
                            1, Split(".", get_endvalue(f"{dbinstance}EngineVersion"))
                        ),
                    ],
                ),
            },
        )


def RDS_DB(key):
    for n, v in getattr(cfg, key).items():
        mapname = f"{key}{n}"

        if not v["IBOX_ENABLED"]:
            continue

        # trick to keep current in use resname and obj.title
        resname = v.get("IBOX_RESNAME", mapname)

        # resources
        r_DB = RDSDBInstance(resname)
        auto_get_props(r_DB, mapname=mapname)
        # trick - when providing DBSnapshotIdentifier
        # or SourceDBInstanceIdentifier some props must not exists
        for m in ["DBName", "MasterUsername", "MasterUserPassword"]:
            setattr(
                r_DB,
                m,
                If(
                    f"{mapname}DBInstanceSkipProperties",
                    Ref("AWS::NoValue"),
                    getattr(r_DB, m),
                ),
            )
        for m in ["SourceDBInstanceIdentifier", "DBSnapshotIdentifier"]:
            setattr(r_DB, m, If(f"{mapname}{m}", getattr(r_DB, m), Ref("AWS::NoValue")))

        # trick fixed name to avoid reboot for now
        # best way should be to use a name like DBParameterGroup{Engine}
        r_PG = RDSDBParameterGroup("DBParameterGroup1", mapname)
        r_PG.Parameters = cfg.DBParameterGroup1

        add_obj([r_DB, r_PG])



import python_minifier
from troposphere import policies, ssm

from .common import *
from .RP import RP_to_cfg

IBOX_SPECIAL_KEYS = (
    "IBOX_RESNAME",
    "IBOX_MAPNAME",
    "IBOX_REMAPNAME",
    "IBOX_INDEXNAME",
    "IBOX_PROPNAME",
    "IBOX_CURNAME",
    "IBOX_REFNAME",
    "IBOX_TITLE",
    "IBOX_LINKED_OBJ_NAME",
    "IBOX_LINKED_OBJ_INDEX",
    "IBOX_LINKED_OBJ_FOR",
)


class IBOX_Custom_Obj(AWSProperty):
    props: PropsDictType = {
        "Value": (str, True),
    }


class Parameter(Parameter):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        if "Type" not in kwargs:
            self.Type = "String"
        if "Default" not in kwargs:
            self.Default = ""

        # Create SSM Parameter for EnvAppXVersion Params to have history of application versions
        if title.startswith("EnvApp") and title.endswith("Version"):
            add_obj(
                [
                    # get_condition(title, "not_equals", "", nomap=True),
                    ssm.Parameter(
                        f"SSMParameter{title}",
                        Type="String",
                        # Condition=title,
                        Name=Sub(
                            "/EnvAppVersions/${EnvRole}/${AWS::StackName}/%s" % title
                        ),
                        Value=Ref(title),
                    ),
                ]
            )


def stack_add_res():
    for n, v in cfg.Parameters.items():
        # Automatically create override conditions for parameters
        if n in list(cfg.fixedvalues) + cfg.mappedvalues:
            if n.endswith("InstanceType"):
                default = "default"
            elif n == "SecurityGroups":
                default = cfg.SECURITY_GROUPS_DEFAULT
            else:
                default = ""

            condition = (
                {f"{n}Override": Not(Equals(Select(0, Ref(n)), default))}
                if v.Type == "CommaDelimitedList"
                else {f"{n}Override": Not(Equals(Ref(n), default))}
            )

            add_obj(condition)
        # End
        cfg.template.add_parameter(v)

    cfg.Parameters.clear()

    for n, v in cfg.Conditions.items():
        cfg.template.add_condition(n, v)
    cfg.Conditions.clear()

    for n, v in cfg.Mappings.items():
        cfg.template.add_mapping(n, v)
    cfg.Mappings.clear()

    for n, v in cfg.Resources.items():
        cfg.template.add_resource(v)
    cfg.Resources.clear()

    for n, v in cfg.Outputs.items():
        add_objoutput(v)
        cfg.template.add_output(v)
    cfg.Outputs.clear()


def add_obj(obj):
    if isinstance(obj, list):
        for n in obj:
            add_obj(n)
    elif isinstance(obj, Parameter):
        cfg.Parameters[obj.title] = obj
    elif isinstance(obj, dict):
        cfg.Conditions.update(obj)
    elif isinstance(obj, Output):
        cfg.Outputs[obj.title] = obj
    else:
        title = obj.title
        if hasattr(obj, "resource_type"):
            # if AWSObject add obj to resources
            add_to = "Resources"
        else:
            # otherway add it to OBJS, usefull to include it later
            add_to = "OBJS"
            if hasattr(obj, "Condition"):
                # Condition is not supported on Resource properties, but use it
                # to wrap obj in an If Condition - needed for LambdaFunctionAssociation
                cond_name = parse_ibox_key(obj.Condition)
                del obj.properties["Condition"]
                obj = If(cond_name, obj, Ref("AWS::NoValue"))

        getattr(cfg, add_to)[title] = obj


def add_objoutput(res):
    try:
        iboxprops = res.IBOX_PROPS
    except Exception:
        pass
    else:
        mapname = iboxprops["MAP"][res.title]

        if isinstance(res.Value, str):
            join_list = []
            for n in res.Value.split():
                n = n.strip()
                if n.startswith("${"):
                    n = n.strip("${}")
                    obj = iboxprops[f"{mapname}{n}"][0]
                    propname = iboxprops[f"{mapname}{n}"][1]
                    n = getattr(obj, propname)
                elif n.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    n = eval(n, globals(), {"IBOX_MAPNAME": mapname})
                join_list.append(n)

            res.Value = Join("", join_list)
        else:
            # output use relative resource condition if do not have one
            # i do this only if Value is not a string
            # i need Condition only if using Ref for getting the resource
            try:
                condition = res.Condition
            except Exception:
                try:
                    # look for condition on current-resource
                    res.Condition = iboxprops[f"{mapname}Condition"][0].Condition
                except Exception:
                    # look for condition on root-resource
                    if hasattr(iboxprops["OBJ"], "Condition"):
                        res.Condition = iboxprops["OBJ"].Condition
            else:
                # Use a fake Condition wih value None in cfg to avoid creating a condition
                if condition == "None":
                    del res.properties["Condition"]

        del res.properties["IBOX_PROPS"]


def do_no_override(action):
    if action is True:
        cfg.no_override = True
    else:
        cfg.no_override = False


def get_endvalue(
    param,
    ssm=False,
    condition=False,
    nocondition=False,
    nolist=False,
    inlist=False,
    split=False,
    issub=False,
    strout=False,
    fixedvalues=None,
    mapinlist=False,
    no_check_for_mappedvalues=False,
    add_override=True,
):
    if not fixedvalues:
        # set default if not defined
        fixedvalues = cfg.fixedvalues

    def _get_overridevalue(param, value, condname=None):
        if param not in cfg.Parameters and condname in cfg.Parameters:
            param = condname
        if (
            cfg.no_override is False
            and add_override
            and param in cfg.Parameters
            and param in list(cfg.fixedvalues) + cfg.mappedvalues
        ):
            override_value = If(f"{param}Override", Ref(param), value)
        else:
            override_value = value

        return override_value

    def _get_value():
        # if param in fixedvalues means its value do not changes
        # based on Env/Region so hardcode the value.
        if param in fixedvalues:
            value = fixedvalues[param]
            # check if value start with method and use eval to run code
            if isinstance(value, list):
                value = [
                    eval(r) if str(r).startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else r
                    for r in value
                ]
            if isinstance(value, str):
                value = (
                    eval(value.replace("\n", ""))
                    if value.startswith(cfg.EVAL_FUNCTIONS_IN_CFG)
                    else value
                )
        # ... otherway use mapping
        elif no_check_for_mappedvalues or param in cfg.mappedvalues:
            value = FindInMap(Ref("EnvShort"), Ref("AWS::Region"), param)
        # .. or return relative paramter
        elif param in cfg.Parameters:
            return Ref(param)
        else:
            raise ValueError(f"{param} not present in mappedvalues.")

        if strout is True and isinstance(value, int):
            value = str(value)

        if nolist is True and isinstance(value, list):
            value = ",".join(value)

        if issub:
            value = Sub(value)

        return value

    value = _get_value()

    if mapinlist is not False:
        value = Select(mapinlist, value)

    if condition or nocondition:
        if condition is True or nocondition is True:
            condname = param
        elif condition:
            condname = condition
        elif nocondition:
            condname = nocondition

        value = _get_overridevalue(param, value, condname)

        value = If(
            condname,
            value if condition else Ref("AWS::NoValue"),
            Ref("AWS::NoValue") if condition else value,
        )
    else:
        value = _get_overridevalue(param, value)

    if split is not False:
        value = Select(split, Split(",", value))

    if inlist is not False:
        value = Select(inlist, value)

    return value


def get_expvalue(param, stack=False, prefix=""):
    v = ""
    if stack:
        v = ImportValue(
            Sub("%s-${%s}" % (param, stack), **{stack: get_endvalue(stack)})
        )
    elif not isinstance(param, str):
        v = ImportValue(Sub("%s${ImportName}" % prefix, **{"ImportName": param}))
    else:
        v = ImportValue(param)

    return v


def get_subvalue(substring, subvar, stack=False):
    submap = {}
    found = substring.find("${")

    while found != -1:
        posindex = found + 2
        myindex = substring[posindex]
        mytype = substring[posindex + 1]
        # M = Mapped, E = Exported
        if myindex.isdigit() and mytype in ["M", "E"]:
            listitem = subvar[int(myindex) - 1] if isinstance(subvar, list) else subvar
            stackitem = stack[int(myindex) - 1] if isinstance(stack, list) else stack
            if mytype == "M":
                submap[listitem] = get_endvalue(listitem)
            else:
                submap[listitem] = get_expvalue(listitem, stackitem)

            substring = substring.replace(
                "${%s%s}" % (myindex, mytype), "${%s}" % listitem
            )

        found = substring.find("${", posindex)

    v = Sub(substring, **submap)

    return v


def get_resvalue(resname, propname):
    res = cfg.Resources[resname]

    loc = propname.find(".")
    while loc > 0:
        prop = propname[0:loc]
        res = getattr(res, prop)
        propname = propname[loc + 1 :]
        loc = propname.find(".")

    return getattr(res, propname)


def iboxif(if_wrapper, mapname, value):
    condname = if_wrapper[0].replace("IBOX_MAPNAME", mapname)
    condname = parse_ibox_key(condname)
    condvalues = []
    for i in if_wrapper[1:3]:
        if isinstance(i, str) and i.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
            v = eval(i)
        else:
            v = i
        condvalues.append(v)

    if condvalues[0] == "IBOX_IFVALUE":
        value = If(condname, value, condvalues[1])
    else:
        value = If(condname, condvalues[0], value)

    return value


def get_dictvalue(key):
    if isinstance(key, list):
        value = [get_dictvalue(k) for k in key]
    elif isinstance(key, dict) and "IBOX_LIST" in key:
        # Usefull for KMS policy and other generic dict properties
        value = []
        for i, k in copy.deepcopy(key).items():
            if i == "IBOX_LIST":
                continue
            # parse IBOX_IF
            if_wrapper = k.pop("IBOX_IF", [])
            prop_obj = {j: get_dictvalue(w) for j, w in k.items()}
            if if_wrapper:
                value.append(iboxif(if_wrapper, "", prop_obj))
            else:
                value.append(prop_obj)
    elif isinstance(key, dict):
        value = {i: get_dictvalue(k) for i, k in key.items()}
    elif isinstance(key, str):
        value = eval(key) if key.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else key
    else:
        value = key

    return value


def get_condition(
    cond_name, cond, value2, key=None, OrExtend=[], mapinlist=False, nomap=None
):
    # record current state
    override_state = cfg.no_override
    do_no_override(True)

    key_name = key if key else cond_name
    if isinstance(key, FindInMap):
        map_name = key.data["Fn::FindInMap"][0]
        key_name = key.data["Fn::FindInMap"][1]
        value_name = key.data["Fn::FindInMap"][2]
        if not value_name and cond_name:
            value_name = cond_name

        value1_param = FindInMap(map_name, Ref(key_name), value_name)
        value1_map = FindInMap(map_name, get_endvalue(key_name), value_name)
    elif isinstance(key, Select):
        select_index = key.data["Fn::Select"][0]
        select_list = key.data["Fn::Select"][1]

        if "Fn::Split" in select_list.data:
            split_sep = select_list.data["Fn::Split"][0]
            key_name = select_list.data["Fn::Split"][1]
            select_value_param = Split(split_sep, Ref(key_name))
            select_value_map = Split(split_sep, get_endvalue(key_name))
        else:
            select_value_param = select_list
            select_value_map = get_endvalue(select_list)

        value1_param = Select(select_index, select_value_param)
        value1_map = Select(select_index, select_value_map)
    else:
        value1_param = Ref(key_name)
        # Used new param "mapinlist" when you have a mapped value in a list
        # but multiple values as override parameters
        if mapinlist:
            value1_map = get_endvalue(mapinlist[0], mapinlist=mapinlist[1])
        elif nomap:
            # if nomap need to avoid checking that value does exist in mappedvalues
            value1_map = get_endvalue(key_name, no_check_for_mappedvalues=True)
        else:
            value1_map = get_endvalue(key_name)

    # if beginning state was False set it back
    if not override_state:
        do_no_override(False)

    eq_param = Equals(value1_param, value2)
    eq_map = Equals(value1_map, value2)

    if cond == "equals":
        cond_param = eq_param
        cond_map = eq_map
    elif cond == "not_equals":
        cond_param = Not(eq_param)
        cond_map = Not(eq_map)

    if (
        key_name in cfg.Parameters
        and key_name in list(cfg.fixedvalues) + cfg.mappedvalues
        and not nomap
    ):
        key_override = f"{key_name}Override"
        condition = Or(
            And(
                Condition(key_override),
                cond_param,
            ),
            And(
                Not(Condition(key_override)),
                cond_map,
            ),
        )
        if OrExtend:
            condition.data["Fn::Or"].extend(OrExtend)
    elif nomap:
        condition = cond_param
    else:
        condition = cond_map

    if cond_name:
        condition = {cond_name: condition}

    return condition


def import_user_data(name):
    TK_IN_UDATA = "_IBOX_CODE_"
    parent_dir_name = os.path.dirname(os.path.realpath(__file__))

    udata_file_main = os.path.join(parent_dir_name, f"user-data/{name}.sh")
    udata_file_ext = os.path.join(os.getcwd(), f"lib/user-data/{name}.sh")

    if os.path.basename(name).islower() and os.path.exists(udata_file_ext):
        # for envrole external script
        udata_file = udata_file_ext
    else:
        udata_file = udata_file_main

    if not os.path.exists(udata_file):
        return []

    try:
        with open(udata_file, "r") as f:
            code_lines = f.read().splitlines(keepends=True)

            file_lines = []
            # parse code for Token IBOX CODE
            for x in code_lines:
                if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    value = eval(x)
                elif x.startswith(TK_IN_UDATA):
                    value = '"'
                elif TK_IN_UDATA in x:
                    # parse Token in string
                    tks = x.split(TK_IN_UDATA)
                    file_lines.extend([f"{tks[0]}", eval(tks[1]), tks[2]])
                    continue
                else:
                    value = "".join(x)

                file_lines.append(value)

            return file_lines

    except IOError:
        raise
    except Exception as e:
        logging.error(f"Error importing user-data {name}: {e}")
        exit(1)


def import_lambda(name):
    TK_IN_LBD = "IBOX_CODE_IN_LAMBDA"
    parent_dir_name = os.path.dirname(os.path.realpath(__file__))
    lambda_file = os.path.join(os.getcwd(), "lib/lambdas/%s.code" % name)
    if not os.path.exists(lambda_file):
        lambda_file = os.path.join(parent_dir_name, "lambdas/%s.code" % name)
    try:
        lambda_file_trunk = os.path.join(*PurePath(lambda_file).parts[-3:])
        with open(lambda_file, "r") as f:
            fdata = f.read()

            try:
                # try to minify using python_minifier
                code = python_minifier.minify(fdata)
                if len(code) > 4096:
                    logging.warning(
                        f"{lambda_file_trunk} > 4096, trying to minify it using a more aggressive option [rename_globals=True]"
                    )
                    code = python_minifier.minify(
                        fdata, rename_globals=True, preserve_globals=["lambda_handler"]
                    )
            except Exception as e:
                # logging.error(f"Failed minifying: {e}")
                code = fdata

            if len(code) > 4096:
                logging.warning(f"Inline lambda {lambda_file_trunk} {len(code)} > 4096")

            code_lines = code.splitlines(keepends=True)

            file_lines = []
            # parse lambda code for Token IBOX CODE
            for x in code_lines:
                if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    value = eval(x)
                elif x.startswith(TK_IN_LBD):
                    value = '"'
                elif TK_IN_LBD in x:
                    # parse minified code
                    tks = x.split(TK_IN_LBD)
                    file_lines.extend([f"{tks[0]}", eval(tks[1]), tks[2]])
                    continue
                else:
                    value = "".join(x)

                file_lines.append(value)

            return file_lines

    except IOError:
        logging.warning(f"Lambda code {name} not found")
        exit(1)
    except Exception as e:
        logging.error(e)


def auto_get_props(
    obj,
    mapname="",
    key=None,
    rootdict=None,
    indexname="",
    remapname="",
    linked_obj_name="",
    linked_obj_index="",
    res_obj_type=None,
):
    # IBOX_RESNAME can be used in yaml and resolved inside get_endvalue
    global IBOX_RESNAME, IBOX_MAPNAME, IBOX_INDEXNAME, IBOX_REMAPNAME, IBOX_PROPNAME
    global IBOX_LINKED_OBJ_NAME, IBOX_LINKED_OBJ_INDEX
    IBOX_RESNAME = obj.title
    IBOX_PROPNAME = ""
    if not "IBOX_MAPNAME" in globals() or mapname:
        IBOX_MAPNAME = mapname
    if not "IBOX_REMAPNAME" in globals() or remapname:
        IBOX_REMAPNAME = remapname
    if not "IBOX_INDEXNAME" in globals() or indexname:
        IBOX_INDEXNAME = indexname
    if not "IBOX_LINKED_OBJ_NAME" in globals() or linked_obj_name:
        IBOX_LINKED_OBJ_NAME = linked_obj_name
    if not "IBOX_LINKED_OBJ_INDEX" in globals() or linked_obj_index:
        IBOX_LINKED_OBJ_INDEX = linked_obj_index

    # create a dict where i will put all property with a flat hierarchy
    # with the name equals to the mapname and the relative value.
    # Later i will assign this dict to the relative output object using
    # IBOX_OUTPUT
    if not "IBOX_PROPS" in globals():
        IBOX_PROPS = {"MAP": {}, "OBJ": obj}

    res_obj_type = getattr(obj, "resource_type", res_obj_type)

    def _get_obj(obj, key, obj_propname, mapname):
        props = obj.props
        mapname_obj = f"{mapname}{obj_propname}"
        global IBOX_PROPNAME, IBOX_REFNAME

        def _get_obj_tags():
            prop_list = []
            for k, v in key[obj_propname].items():
                try:
                    v["Value"]
                except Exception:
                    pass
                else:
                    v["Value"] = get_endvalue(f"{mapname_obj}{k}Value")

                try:
                    if_wrapper = v["IBOX_IF"]
                except Exception:
                    pass
                else:
                    del v["IBOX_IF"]
                    v = iboxif(if_wrapper, mapname, v)

                prop_list.append(v)

            obj_tags = Tags()
            obj_tags.tags = prop_list

            return obj_tags

        obj_class = obj.__class__.__name__
        obj_mod_name = obj.__module__
        prop_class = props.get(obj_propname, [None])[0]
        # print(f"TYPE: {res_obj_type}, CLASS: {obj_class}, PROP: {obj_propname}")

        # detect attributes as CreationPolicy not included in obj properties
        if obj_propname in ["CreationPolicy", "UpdatePolicy"]:
            prop_class = getattr(policies, obj_propname)
        elif callable(prop_class) and prop_class.__name__ not in [
            "validate_variables_name",
            "policytypes",
        ]:
            # prop_class is a method fallback and try to use CloudFormationResourceSpecification
            if res_obj_type in cfg.TROPO_CLASS_TO_CFM:
                obj_class = cfg.TROPO_CLASS_TO_CFM[res_obj_type].get(
                    obj_class, obj_class
                )
            try:
                res_obj_propname = cfg.cfm_res_spec["PropertyTypes"][
                    f"{res_obj_type}.{obj_class}"
                ]["Properties"][obj_propname]
            except Exception:
                try:
                    res_obj_propname = cfg.cfm_res_spec["ResourceTypes"][res_obj_type][
                        "Properties"
                    ][obj_propname]
                except Exception:
                    res_obj_propname = None

            if res_obj_propname:
                obj_prop_type = res_obj_propname.get("Type")
                prop_class_type = res_obj_propname.get("ItemType")
                if not prop_class_type:
                    prop_class_type = obj_prop_type
                if not prop_class_type:
                    prop_class_type = res_obj_propname.get("PrimitiveType")

                # print(f"PROP_CLASS_TYPE: {prop_class_type}")
                try:
                    prop_class = getattr(sys.modules[obj_mod_name], prop_class_type)
                except Exception:
                    pass
                else:
                    if obj_prop_type == "List":
                        prop_class = [prop_class]

        # print(f"PROP_CLASS: {prop_class}")

        # obj_propname is a class, usually Tropo AWSProperty
        if isinstance(prop_class, type):
            # If object already have that props, object class is
            # the existing already defined object,
            # so extend its property with auto_get_props
            if obj_propname in obj.properties:
                prop_obj = obj.properties[obj_propname]
            else:
                prop_obj = prop_class()

            if prop_class.__bases__[0].__name__ in ["AWSProperty", "AWSAttribute"]:
                _populate(
                    prop_obj,
                    key=key[obj_propname],
                    mapname=mapname_obj,
                )
                # Check for incomplete AWSProperty object and set obj to None to skip it
                try:
                    prop_obj.to_dict()
                except Exception as e:
                    # logging.warning(f"Resource with missing properties: {obj_propname}\n\t\t{e}")
                    prop_obj = None
            elif prop_class.__name__ == "dict":
                prop_obj = get_dictvalue(key[obj_propname])
            elif prop_class.__name__ == "Tags":
                prop_obj = _get_obj_tags()
            elif prop_class.__name__ == "str" and obj_propname == "LifecyclePolicyText":
                # str but can be represented as dict Ex. ECR LifecyclePolicyText
                prop_obj = json.dumps(get_dictvalue(key[obj_propname]))

            return prop_obj

        # obj_propname is are Tropo Tags
        elif (
            isinstance(prop_class, tuple)
            and any(n in prop_class for n in [Tags, asgTags])
        ) or (
            callable(prop_class) and prop_class.__name__ in ["validate_tags_or_list"]
        ):
            prop_obj = _get_obj_tags()

            return prop_obj

        # obj_propname is List of Tropo AWSProperty
        elif (
            isinstance(prop_class, list)
            and isinstance(prop_class[0], type)
            and prop_class[0].__bases__[0].__name__ == "AWSProperty"
        ):
            prop_list = []
            prop_class = prop_class[0]

            # save IBOX_REFNAME
            if "IBOX_REFNAME" in globals():
                globals()[f"IBOX_REFNAME@{mapname}"] = globals()["IBOX_REFNAME"]

            for o, v in key[obj_propname].items():
                # for a list of properties set IBOX_PROPNAME to the name of property
                IBOX_PROPNAME = o

                if o == "IBOX_IF":
                    # element named IBOX_IF must no be parsed
                    # is needed for wrapping whole returned obj in _populate
                    continue
                name_o = str(o)
                mapname_o = f"{mapname_obj}{name_o}"
                prop_obj = prop_class()

                _populate(prop_obj, key=v, mapname=mapname_o)

                # trick to wrapper single obj in If Condition
                try:
                    if_wrapper = v["IBOX_IF"]
                except Exception:
                    prop_list.append(prop_obj)
                else:
                    prop_list.append(iboxif(if_wrapper, mapname, prop_obj))

            # restore IBOX_REFNAME
            if f"IBOX_REFNAME@{mapname}" in globals():
                IBOX_REFNAME = globals()[f"IBOX_REFNAME@{mapname}"]

            if prop_list:
                return prop_list

        # obj_propname is a dict as KSM Key KeyPolicy
        elif (
            isinstance(prop_class, tuple)
            and isinstance(prop_class[0], type)
            and prop_class[0].__name__ == "dict"
        ) or (
            callable(prop_class)
            and prop_class.__name__ in ["policytypes", "validate_variables_name"]
        ):
            return get_dictvalue(key[obj_propname])

    def _populate(obj, key=None, mapname=None, rootdict=None):
        global IBOX_RESNAME, IBOX_CURNAME, IBOX_REFNAME, IBOX_TITLE

        if not mapname:
            mapname = obj.title
        if rootdict:
            key = rootdict
            mapname = ""
        if not key:
            key = getattr(cfg, mapname)

        if key.get("IBOX_TITLE"):
            IBOX_TITLE = parse_ibox_key(key["IBOX_TITLE"])

        def _try_PCO_in_obj(key):
            def _parameter(k):
                for n, v in k.items():
                    if not eval(v.get("IBOX_ENABLED_IF", "True")):
                        continue
                    n = parse_ibox_key(n)
                    parameter = Parameter(n)
                    _populate(parameter, rootdict=v)
                    add_obj(parameter)

            def _condition(k):
                for n, v in k.items():
                    n = parse_ibox_key(n)
                    condition = {n: eval(v)}
                    add_obj(condition)

            def _output(k):
                for n, v in k.items():
                    n = parse_ibox_key(n)
                    if not eval(v.get("IBOX_ENABLED_IF", "True")):
                        continue
                    output = Output(n)
                    _populate(output, rootdict=v)
                    # alter troposphere obj and add IBOX property
                    output.props["IBOX_PROPS"] = (dict, False)
                    # starting from troposphere 3.0 propnames is a set
                    output.propnames = list(output.propnames) + ["IBOX_PROPS"]
                    # assign auto_get_props populated obj to IBOX property
                    output.IBOX_PROPS = IBOX_PROPS
                    output.IBOX_PROPS["MAP"][n] = mapname
                    add_obj(output)

            func_map = {
                "IBOX_PARAMETER": _parameter,
                "IBOX_CONDITION": _condition,
                "IBOX_OUTPUT": _output,
            }
            for pco in list(func_map.keys()):
                try:
                    k = key[pco]
                except Exception:
                    pass
                else:
                    func_map[pco](k)

        def _auto_PO(name, conf, mode, value=""):
            conf = dict(conf)
            if "p" in mode:
                parameter_base = {"Description": "Empty for mapped value"}
                parameter_base.update(conf)
                # avoid including Output Condition in Parameter
                parameter_base.pop("Condition", "")
                pco_conf = {"IBOX_PARAMETER": {f"{mapname}{name}": parameter_base}}
                # look for CONDITION key and if found update pco_conf
                condition = conf.get("CONDITION")
                if condition:
                    pco_conf.update({"IBOX_CONDITION": {f"{mapname}{name}": condition}})
                _try_PCO_in_obj(pco_conf)
            if "o" in mode:
                # avoid parsing Parameter Descrption as output one
                conf.pop("Description", "")
                if isinstance(value, int):
                    # Output value must be a string
                    value = str(value)
                elif isinstance(value, str) and value.startswith("get_endvalue"):
                    value = eval(value)
                output_base = {"Value": value}
                output_base.update(conf)
                if output_base["Value"] != "IBOX_SKIP":
                    output = {"IBOX_OUTPUT": {f"{mapname}{name}": output_base}}
                    _try_PCO_in_obj(output)

        def _process_ibox_auto_pco_key(propname):
            ibox_auto_p = f"{propname}.IBOX_AUTO_P"
            ibox_auto_po = f"{propname}.IBOX_AUTO_PO"
            ibox_pco = f"{propname}.IBOX_PCO"

            # IBOX_AUTO_PO
            for n in [ibox_auto_po, ibox_auto_p]:
                if n in key:
                    # Automatically create parameter
                    _auto_PO(propname, key[n], "p")

            # IBOX_PCO
            if ibox_pco in key:
                # If there is a key ending with {propname}.IBOX_PCO process it
                _try_PCO_in_obj(key[ibox_pco])

        def _proc_custom_obj(base_rootdict, key_value, mapname, propname):
            global IBOX_RESNAME, IBOX_MAPNAME

            values = []
            if not base_rootdict:
                base_rootdict = getattr(cfg, f"{mapname}IBOX_CUSTOM_OBJ{propname}")
            if isinstance(key_value, dict):
                my_iter = iter(key_value.items())
                is_dict = True
            else:
                my_iter = enumerate(key_value)
                is_dict = False

            for n, v in my_iter:
                rootdict = {"Value": v}
                obj_title = n if is_dict else v
                if isinstance(v, dict):
                    rootdict.update(v.get("Conf", {}))
                    v = v.get("Value", v)
                    obj_title = f"{n}Value"
                rootdict.update(base_rootdict)
                obj = IBOX_Custom_Obj(obj_title)

                # save IBOX_RESNAME and IBOX_MAPNAME
                IBOX_RESNAME_SAVE = IBOX_RESNAME
                IBOX_MAPNAME_SAVE = IBOX_MAPNAME
                # set IBOX_RESNAME and IBOX_MAPNAME to call _populate like auto_get_props
                IBOX_RESNAME = obj_title
                IBOX_MAPNAME = f"{mapname}{propname}"
                # populate OBJ
                _populate(obj, rootdict=rootdict, mapname=IBOX_MAPNAME)
                # restore IBOX_RESNAME and IBOX_MAPNAME
                IBOX_RESNAME = IBOX_RESNAME_SAVE
                IBOX_MAPNAME = IBOX_MAPNAME_SAVE

                value = getattr(obj, "Value")
                if v == "IBOX_IFCONDVALUE":
                    condname = f"{mapname}{propname}{n}"
                    add_obj(get_condition(condname, "not_equals", "IBOX_IFCONDVALUE"))
                    values.append(If(condname, value, Ref("AWS::NoValue")))
                else:
                    values.append(value)
            return values

        def _linked_obj(linked_obj_data):
            # need to parse them here to have the right values
            lo_name = parse_ibox_key(linked_obj_data.get("Name", IBOX_RESNAME))
            lo_key = parse_ibox_key(linked_obj_data.get("Key", ""))
            lo_type = parse_ibox_key(linked_obj_data.get("Type", ""))
            # this way even if without key "For", for cycle will run at least one time
            lo_for_cycle = linked_obj_data.get("For", [""])

            louc_cfg = {}
            for lo_for_index in lo_for_cycle:
                # need to copy it because it's updated
                lo_conf = copy.deepcopy(
                    linked_obj_data.get("Conf", {"IBOX_LINKED_OBJ_NAME": IBOX_RESNAME})
                )
                lo_conf["IBOX_ENABLED"] = True
                # for all lo_conf entries, if their value is a string parse it using parse_ibox_key
                for loc_entry_key, loc_entry_value in lo_conf.items():
                    lo_conf[loc_entry_key] = parse_ibox_key(loc_entry_value)
                    lo_conf[loc_entry_key] = parse_ibox_key(
                        loc_entry_value, conf={"IBOX_LINKED_OBJ_FOR": lo_for_index}
                    )

                linked_obj_key_name = f"{lo_key}{lo_type}{lo_for_index}"
                target_name = f"{lo_name}{lo_for_index}"

                # get existing object
                if f"{lo_key}{target_name}" == linked_obj_key_name:
                    # source and target are the same, so update the object in place
                    linked_obj = getattr(cfg, linked_obj_key_name)
                else:
                    # copy it, first search for the full name including for index
                    linked_obj = copy.deepcopy(
                        getattr(
                            cfg,
                            linked_obj_key_name,
                            getattr(cfg, f"{lo_key}{lo_type}", None),
                        )
                    )

                # update it with config from IBOX_LINKED_OBJ Conf
                linked_obj.update(lo_conf)

                lo_resname = lo_conf.get("IBOX_RESNAME", f"{lo_key}{lo_name}")
                linked_obj["IBOX_RESNAME"] = f"{lo_resname}{lo_for_index}"

                condition = key.get("Condition", getattr(obj, "Condition", None))
                if condition and not "Condition" in linked_obj:
                    # automatically add Condition if source obj have it and target not
                    linked_obj["Condition"] = parse_ibox_key(condition)

                # assign to louc_cfg lo_key
                louc_cfg[target_name] = linked_obj

            if cfg.debug:
                pprint(louc_cfg)
            # update cfg and fixedvalues, need to do it this way to avoid overwriting the lo_key and removing all objects
            RP_to_cfg(louc_cfg, prefix=lo_key, mappedvalues=cfg.mappedvalues)
            # finally update cfg object base key with configuration including mutiple objects
            getattr(cfg, lo_key).update(louc_cfg)

        # Allowed object properties
        props = list(vars(obj)["propnames"])
        props.extend(vars(obj)["attributes"])

        # needed by IBOX_BASE used on Resource Properties
        if key.get("IBOX_BASE_REF"):
            IBOX_REFNAME = mapname

        # Parameters, Conditions, Outpus in yaml cfg
        _try_PCO_in_obj(key)

        # For linked resources as r53recordset, update their keys conf.
        # This way they can be enabled or their name can be changed, or so on...
        ibox_linked_obj = key.get("IBOX_LINKED_OBJ")
        if ibox_linked_obj:
            # bad hack to use a dict for ibox_linked_obj, so that i can modify it if needed using the index name
            # but avoid intercepting the "normal" conf that is a dict too (a key named Type must not exist)
            if isinstance(ibox_linked_obj, dict) and "Type" not in ibox_linked_obj:
                for linked_obj in ibox_linked_obj.values():
                    if linked_obj:
                        _linked_obj(linked_obj)
            else:
                _linked_obj(ibox_linked_obj)

        # iterate over both obj props and key - using key order
        for propname in dict(
            list(key.items()) + list(dict.fromkeys(props, None).items())
        ).keys():
            if key.get(propname) == "IBOX_SKIP":
                continue

            # IBOX KEY TO LOOK FOR
            ibox_auto_p = f"{propname}.IBOX_AUTO_P"
            ibox_auto_po = f"{propname}.IBOX_AUTO_PO"
            ibox_pco = f"{propname}.IBOX_PCO"
            ibox_code = f"{propname}.IBOX_CODE"
            ibox_code_key = f"{propname}.IBOX_CODE_KEY"
            ibox_custom_obj = f"{propname}.IBOX_CUSTOM_OBJ"
            ibox_propname_fixed = f"{propname}@"

            # need to make work @ override even if propname is not in mapped values
            if ibox_propname_fixed in key and propname not in cfg.mappedvalues:
                key[propname] = key[ibox_propname_fixed]
                cfg.fixedvalues[f"{mapname}{propname}"] = key[propname]

            custom_key_only = propname.replace(".IBOX_AUTO_PO", "")

            if not isinstance(obj, (Output, Parameter, Condition)):
                IBOX_CURNAME = f"{mapname}{propname}"

            # IBOX_PCO and IBOX_AUTO_PO for Custom Key ONLY
            if (
                ibox_pco in key
                and propname in key
                and propname not in props
                and all(n not in key for n in [ibox_auto_p, ibox_auto_po, ibox_code])
            ):
                # process ibox_pco for custom key not present in obj props
                # but only if there is not a relative ibox_auto_p/o or ibox_code key
                _try_PCO_in_obj(key[ibox_pco])
            elif propname.endswith("IBOX_AUTO_PO") and custom_key_only not in props:
                # process IBOX_AUTO_PO custom key if the relative propname do not exist in resource props
                _auto_PO(
                    custom_key_only,
                    key[propname],
                    "po",
                    f"get_endvalue('{mapname}{custom_key_only}')",
                )

            # IBOX_CODE
            if ibox_code in key and propname in props:
                # If there is a key ending with {prop}.IBOX_CODE  eval it and use it as prop value.
                # First process IBOX_AUTO_PO and IBOX_PCO keys
                _process_ibox_auto_pco_key(propname)
                value = eval(key[ibox_code])
            elif (
                propname in key
                and propname in props
                or (propname in key and ibox_custom_obj in key)
            ):
                # there is match between obj prop and a dict key or there is a custom obj
                key_value = key[propname]

                # Process IBOX_AUTO_PO and IBOX_PCO keys
                _process_ibox_auto_pco_key(propname)

                # IBOX_CODE_KEY - like IBOX_CODE but only if propname is in key too
                if ibox_code_key in key:
                    value = eval(key[ibox_code_key])
                # set value
                elif isinstance(key_value, (list, dict)) and ibox_custom_obj in key:
                    # to process a list like a custom obj
                    value = _proc_custom_obj(
                        key[ibox_custom_obj], key_value, mapname, propname
                    )
                elif isinstance(key_value, dict):
                    # key value is a dict, get populated object
                    value = _get_obj(obj, key, propname, mapname)
                    if value is None:
                        continue
                elif isinstance(key_value, str) and key_value.startswith(
                    IBOX_SPECIAL_KEYS
                ):
                    # parse value for IBOX special keys
                    value = parse_ibox_key(key_value)
                elif rootdict:
                    # rootdict is needed by lib/efs.py EFS_FileStorage SGIExtra
                    # where is passed as a dictionary to parse for parameters
                    value = get_endvalue(f"{mapname}{propname}", fixedvalues=rootdict)
                elif isinstance(key_value, str) and key_value.startswith(
                    "get_endvalue("
                ):
                    # Usefull to migrate code in yaml using auto_get_props
                    # get_endvalue is used only when migrating code
                    value = eval(key_value)
                else:
                    value = get_endvalue(f"{mapname}{propname}")

                if key_value == "IBOX_IFCONDVALUE":
                    # auto add condition and wrap value in AWS If Condition
                    add_obj(
                        get_condition(
                            f"{mapname}{propname}", "not_equals", "IBOX_IFCONDVALUE"
                        )
                    )
                    value = If(f"{mapname}{propname}", value, Ref("AWS::NoValue"))

                # trick to wrapper recursed returned value in If Condition
                try:
                    if_wrapper = key_value["IBOX_IF"]
                except Exception:
                    pass
                else:
                    value = iboxif(if_wrapper, mapname, value)

                if propname == "Condition" and not isinstance(value, str):
                    # Avoid intercepting a Template Condition
                    # as a Resource Condition
                    continue
            else:
                # NO match between propname and cfg keys
                continue

            # Finally set obj property
            try:
                setattr(obj, propname, value)
            except AttributeError:
                setattr(cfg, f"{mapname}{propname}", value)
                pass
            except TypeError:
                pass
            else:
                # populate IBOX_PROPS dict
                IBOX_PROPS[f"{mapname}{propname}"] = (obj, propname)

                # Automatically create output
                if ibox_auto_po in key:
                    ibox_auto_po_value = key[ibox_auto_po]
                    _auto_PO(propname, ibox_auto_po_value, "o", value)

        # need to redefine it here because it's has been changed by nested supprop
        if not isinstance(obj, (Output, Parameter, Condition)):
            IBOX_CURNAME = mapname

        # title override
        try:
            obj.title = parse_ibox_key(key["IBOX_TITLE"])
        except Exception:
            pass

    _populate(obj, key, mapname, rootdict)
    return obj


def auto_build_obj(obj, key, name=None):
    props = vars(obj)["propnames"]
    classname = obj.__class__
    for resname, resvalue in key.items():
        final_obj = classname(resname)
        if not name:
            name = final_obj.__class__.__name__
        auto_get_props(final_obj, f"{name}{resname}")

        add_obj(final_obj)


def change_obj_data(obj, find, value):
    if isinstance(obj, list):
        for n, v in enumerate(obj):
            change_obj_data(obj[n], find, value)
    elif isinstance(obj, If):
        obj_if = obj.data["Fn::If"]
        for n, v in enumerate(obj_if):
            if isinstance(v, Ref) and v.data["Ref"] == find:
                obj_if[n] = value


def gen_random_string():
    length = 16
    char_set = f"{string.ascii_letters}{string.digits}"
    if not hasattr(gen_random_string, "rng"):
        # Create a static variable
        gen_random_string.rng = random.SystemRandom()
    secret_string = "".join(
        [gen_random_string.rng.choice(char_set) for _ in range(length)]
    )

    return secret_string


def clf_compute_order(pattern):
    base_ord = 1

    for s, w in cfg.CLF_PATH_PATTERN_REPLACEMENT.items():
        pattern = pattern.replace(w, s)

    n_star = 0
    for n, v in enumerate(pattern):
        if v == "?":
            base_ord += 0.0001
        elif v == "*":
            if len(pattern) == n + 1 and n_star != 0:
                # v is last char but not the only star char
                base_ord += 0.0005
            elif n_star != 0:
                base_ord -= 0.000001 * n
            else:
                base_ord += 1000 / n
                n_star += 1
        elif n_star != 0:
            base_ord -= 0.001

    cfg.dbg_clf_compute_order[pattern] = base_ord

    return base_ord


def parse_ibox_key(value, conf={}):
    if not isinstance(value, str):
        return value
    if value.startswith(cfg.EVAL_PYTHON_FUNCTIONS_IN_CFG):
        return eval(value)
    for key in IBOX_SPECIAL_KEYS:
        if key in value:
            if key in conf:
                value = value.replace(key, conf[key])
            else:
                value = value.replace(key, globals().get(key, ""))
    value = value.replace("_", IBOX_RESNAME)
    value = value.replace(".", "")

    return value


def camel_to_snake(data):
    out = ""
    skip_next = False
    for n, v in enumerate(data):
        if skip_next:
            skip_next = False
            continue
        if v.isupper() and data[n + 1].isupper():
            out += f"_{v}{data[n+1]}"
            skip_next = True
        elif v.isupper():
            out += f"_{v}" if out else v
        else:
            out += v

    return out.upper()



import sys
import os
import copy
from pprint import pprint, pformat

from . import cfg


def get_envregion_mapping():
    RP = copy.deepcopy(cfg.RP_map)

    for env in list(RP):
        rvalue = RP[env]
        for region in list(rvalue.keys()):
            for key in cfg.mappedvalues:
                if key not in RP[env][region]:
                    # get base/global value
                    try:
                        RP[env][region][key] = getattr(cfg, key)
                    except Exception:
                        pass
            # delete empty mappings CF do not like them
            if not RP[env][region]:
                del RP[env][region]
        if not RP[env]:
            del RP[env]

    return RP


def get_ec2_mapping():
    mappings = {}

    #   #START##MAP INSTANCE TYPE EPHEMERAL####
    map_eph = {
        "InstaceEphemeral0": [
            "m3.medium",
            "m3.large",
            "m3.xlarge",
            "m3.2xlarge",
            "c3.large",
            "c3.xlarge",
            "c3.2xlarge",
            "c3.4xlarge",
            "g2.2xlarge",
            "r3.large",
            "r3.xlarge",
            "r3.2xlarge",
            "r3.4xlarge",
            "i2.xlarge",
            "i2.2xlarge",
            "i2.4xlarge",
            "d2.xlarge",
            "d2.2xlarge",
            "d2.4xlarge",
        ],
        "InstaceEphemeral1": [
            "m3.xlarge",
            "m3.2xlarge",
            "c3.large",
            "c3.xlarge",
            "c3.2xlarge",
            "c3.4xlarge",
            "i2.2xlarge",
            "i2.4xlarge",
            "d2.xlarge",
            "d2.2xlarge",
            "d2.4xlarge",
        ],
        "InstaceEphemeral2": [
            "i2.4xlarge",
            "d2.xlarge",
            "d2.2xlarge",
            "d2.4xlarge",
        ],
    }

    mappings["InstanceTypes"] = {}

    for i in cfg.INSTANCE_LIST:
        mappings["InstanceTypes"].update(
            {
                i: {
                    "InstaceEphemeral0": "1"
                    if i in map_eph["InstaceEphemeral0"]
                    else "0",
                    "InstaceEphemeral1": "1"
                    if i in map_eph["InstaceEphemeral1"]
                    else "0",
                    "InstaceEphemeral2": "1"
                    if i in map_eph["InstaceEphemeral2"]
                    else "0",
                }
            }
        )
    #   #END##MAP INSTANCE TYPE EPHEMERAL####

    return mappings


def get_azones_mapping():
    mappings = {}
    mappings["AvabilityZones"] = {}
    AZ = cfg.AZones

    for r in cfg.regions:
        zones = {}
        try:
            nzones = AZ[r]
        except Exception:
            nzones = AZ["default"]

        try:
            nzones = cfg.RP["dev"][r]["AZones"]
        except Exception:
            pass

        for n in range(AZ["MAX"]):
            zones[f"Zone{n}"] = "True" if nzones > n else "False"

        mappings["AvabilityZones"][r] = zones

    return mappings


class Mappings(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            if n == "EnvRegion":
                mapping = get_envregion_mapping()
            if n == "EC2":
                mapping = get_ec2_mapping()
            if n == "AZones":
                mapping = get_azones_mapping()

            cfg.Mappings.update(mapping)



import troposphere.elasticloadbalancing as elb
import troposphere.elasticloadbalancingv2 as elbv2
import troposphere.ec2 as ec2

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)
from .ec2 import SecurityGroupIngressInstanceELBPorts


# Temporary fix for https://github.com/cloudtools/troposphere/issues/1474
elbv2.one_of = my_one_of

# Fix troposphere/elasticloadbalancing.py LBCookieStickinessPolicy is a List and do not use class LBCookieStickinessPolicy
elb.LoadBalancer.props["LBCookieStickinessPolicy"] = (
    [elb.LBCookieStickinessPolicy],
    False,
)
elb.LoadBalancer.props["Listeners"] = ([elb.Listener], False)


def enable_recordset(rtype):
    prefix = f"Route53RecordSet"
    if "External" in cfg.RecordSet:
        if "External" in cfg.LoadBalancer:
            record = getattr(cfg, f"{prefix}{rtype}ExternalLoadBalancerExternal")
        else:
            record = getattr(cfg, f"{prefix}{rtype}ExternalLoadBalancerInternal")
        record["IBOX_ENABLED"] = True
    if "Internal" in cfg.RecordSet:
        if "Internal" in cfg.LoadBalancer:
            record = getattr(cfg, f"{prefix}{rtype}InternalLoadBalancerInternal")
        else:
            record = getattr(cfg, f"{prefix}{rtype}InternalLoadBalancerExternal")
        record["IBOX_ENABLED"] = True


def LB_ElasticLoadBalancingClassicEC2():
    for lb in cfg.LoadBalancer:
        # update SecurityGroupInstancesRules Ingress using Listeners
        for n in getattr(cfg, f"ElasticLoadBalancingLoadBalancer{lb}")["Listeners"]:
            r_SGIInstance = SecurityGroupIngressInstanceELBPorts(
                f"SecurityGroupIngressListeners{n}",
                FromPort=get_endvalue(
                    f"ElasticLoadBalancingLoadBalancer{lb}Listeners{n}InstancePort"
                ),
                SourceSecurityGroupId=Ref("SecurityGroupLoadBalancer"),
                ToPort=get_endvalue(
                    f"ElasticLoadBalancingLoadBalancer{lb}Listeners{n}InstancePort"
                ),
            )
            add_obj(r_SGIInstance)

        # LoadBalancer
        r_LB = elb.LoadBalancer(f"LoadBalancerClassic{lb}")
        auto_get_props(r_LB, mapname=f"ElasticLoadBalancingLoadBalancer{lb}")
        add_obj(r_LB)

        cfg.CloudWatchAlarm[f"Backend{lb}5XX"]["IBOX_ENABLED"] = True


def LB_ElasticLoadBalancingApplicationEC2():
    for lb in cfg.LoadBalancer:
        r_LB = elbv2.LoadBalancer(f"LoadBalancerApplication{lb}")
        auto_get_props(
            r_LB, mapname=f"ElasticLoadBalancingV2LoadBalancerEC2Application{lb}"
        )
        add_obj(r_LB)

        # enable Listeners
        if lb == "External":
            getattr(cfg, f"ElasticLoadBalancingV2ListenerEC2Https{lb}")[
                "IBOX_ENABLED"
            ] = True
        else:
            getattr(cfg, f"ElasticLoadBalancingV2ListenerEC2Http{lb}")[
                "IBOX_ENABLED"
            ] = True
        # enable TargetGroups
        getattr(
            cfg, f"ElasticLoadBalancingV2TargetGroupEC2LoadBalancerApplication{lb}"
        )["IBOX_ENABLED"] = True

        cfg.CloudWatchAlarm[f"TargetEC2{lb}5XX"]["IBOX_ENABLED"] = True

    # update SecurityGroupInstancesRules giving access from LB to Target Ports
    for n, v in cfg.ElasticLoadBalancingV2TargetGroup.items():
        if v.get("IBOX_ENABLED", True):
            r_SGIInstance = SecurityGroupIngressInstanceELBPorts(
                f"SecurityGroupIngressListeners{n}",
                FromPort=get_endvalue(f"ElasticLoadBalancingV2TargetGroup{n}Port"),
                SourceSecurityGroupId=Ref("SecurityGroupLoadBalancer"),
                ToPort=get_endvalue(f"ElasticLoadBalancingV2TargetGroup{n}Port"),
            )
            add_obj(r_SGIInstance)


def LB_ElasticLoadBalancingNetworkEC2():
    for lb in cfg.LoadBalancer:
        r_LB = elbv2.LoadBalancer(f"LoadBalancerNetwork{lb}")
        auto_get_props(
            r_LB, mapname=f"ElasticLoadBalancingV2LoadBalancerEC2Network{lb}"
        )
        add_obj(r_LB)

        # enable Listeners
        getattr(cfg, f"ElasticLoadBalancingV2ListenerEC2TCP{lb}")["IBOX_ENABLED"] = True
        # enable TargetGroups
        getattr(cfg, f"ElasticLoadBalancingV2TargetGroupEC2LoadBalancerNetwork{lb}")[
            "IBOX_ENABLED"
        ] = True

    # update SecurityGroupInstancesRules giving access to Target Ports
    for n, v in cfg.ElasticLoadBalancingV2TargetGroup.items():
        if v.get("IBOX_ENABLED", True):
            r_SGIInstance = SecurityGroupIngressInstanceELBPorts(
                f"SecurityGroupIngressListeners{n}",
                FromPort=get_endvalue(f"ElasticLoadBalancingV2TargetGroup{n}Port"),
                CidrIp="0.0.0.0/0",
                ToPort=get_endvalue(f"ElasticLoadBalancingV2TargetGroup{n}Port"),
            )
            add_obj(r_SGIInstance)


def LB_ElasticLoadBalancingEC2(key):
    if not cfg.LoadBalancer:
        return
    enable_recordset("EC2")

    if cfg.LoadBalancerType == "Classic":
        LB_ElasticLoadBalancingClassicEC2()

    if cfg.LoadBalancerType == "Application":
        LB_ElasticLoadBalancingApplicationEC2()

    if cfg.LoadBalancerType == "Network":
        LB_ElasticLoadBalancingNetworkEC2()



from . import cfg
from .shared import stack_add_res
from . import (
    __version__,
    autoscaling,
    cloudformation,
    cloudfront,
    ec2,
    ecr,
    ecs,
    efs,
    iam,
    joker,
    elasticloadbalancing,
    mappings,
    rds,
    s3,
    waf,
)


def execute_method(RP_cmm):
    processed = []

    def _process(k, v):
        func_name = v["func"]
        module_name = v["module"]
        module = globals()[module_name]
        dep = v.get("dep", [])

        if k in list(RP_cmm.keys()):
            # process dependency modules first
            for m in dep:
                if m in list(RP_cmm.keys()) and m not in processed:
                    _process(m, cfg.CFG_TO_FUNC[m])
                    processed.append(m)

            RP_value = RP_cmm[k]
            if isinstance(RP_value, str) and RP_value == "IBOX_SKIP_FUNC":
                return
            if isinstance(func_name, list):
                for n in func_name:
                    getattr(module, n)(key=k)
                return
            stacktype_func = f"{func_name}{cfg.stacktype.upper()}"
            if stacktype_func in dir(module):
                getattr(module, stacktype_func)(key=k)
            elif func_name in dir(module):
                getattr(module, func_name)(key=k)
            elif module_name == "joker":
                # for resources that can be built using only auto_get_props
                joker.Joker(key=k, module=func_name[0], cls=func_name[1])
            processed.append(k)

    for k, v in cfg.CFG_TO_FUNC.items():
        # skip already processed keys
        if k in processed:
            continue
        _process(k, v)

    stack_add_res()


def generate():
    try:
        execute_method(cfg.RP_tree)
    except Exception:
        # in case of error call stack_add_res to empty the cfg Parameters
        # Conditions Mappings Resources Outputs  dicts. This is needed cause
        # get_template "restore base cfg" section do not recursive remove
        # "sub-objetcs" or to say it in a better way:
        # dict clear methods reset the contents,
        # but all references to the same instance will still be present.
        stack_add_res()
        raise

    version_text = f"(Generated by awsibox >= {__version__})"
    cfg.template.set_description(f"{cfg.envrole} {version_text} [{cfg.stacktype}]")
    cfg.template.set_version("2010-09-09")

    return cfg.template



import troposphere.ecr as ecr

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


def ECRRepositoryPolicyStatementAccountPull(name):
    policy = {
        "Action": [
            "ecr:GetDownloadUrlForLayer",
            "ecr:BatchGetImage",
            "ecr:BatchCheckLayerAvailability",
            "ecr:ListImages",
            "ecr:DescribeRepositories",
            "ecr:DescribeImages",
        ],
        "Effect": "Allow",
        "Principal": {"AWS": [get_subvalue("arn:aws:iam::${1M}:root", name)]},
        "Sid": "AllowPull",
    }

    return policy


def ECRRepositoryPolicyStatementAccountPush(name):
    policy = {
        "Action": [
            "ecr:PutImage",
            "ecr:InitiateLayerUpload",
            "ecr:UploadLayerPart",
            "ecr:CompleteLayerUpload",
            "ecr:BatchCheckLayerAvailability",
        ],
        "Effect": "Allow",
        "Principal": {"AWS": [get_subvalue("arn:aws:iam::${1M}:root", name)]},
        "Sid": "AllowPush",
    }

    return policy


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def ECR_Repositories(key):
    PolicyStatementAccounts = []
    for n, v in cfg.EcrAccount.items():
        mapname = f"EcrAccount{n}Id"  # Ex. EcrAccountPrdId
        # conditions
        add_obj(get_condition(mapname, "not_equals", "none"))

        if "Pull" in v["Policy"]:
            PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPull(
                name=mapname
            )
            PolicyStatementAccounts.append(
                If(mapname, PolicyStatementAccount, Ref("AWS::NoValue"))
            )

        if "Push" in v["Policy"]:
            PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPush(
                name=mapname
            )
            PolicyStatementAccounts.append(
                If(mapname, PolicyStatementAccount, Ref("AWS::NoValue"))
            )

    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"
        Repo = ecr.Repository(resname)
        auto_get_props(Repo, indexname=n)
        Repo.RepositoryPolicyText["Statement"].extend(PolicyStatementAccounts)

        add_obj(Repo)



import troposphere.iam as iam

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


def IAM_Users(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"  # Ex. IAMUserPincoPalla
        if not v.get("IBOX_ENABLED", True):
            continue

        ManagedPolicyArns = []
        RoleGroups = []
        if "RoleGroups" in v:
            for m, w in v["RoleGroups"].items():
                condname = f"{resname}RoleGroups{m}"
                # conditions
                add_obj(get_condition(condname, "equals", "yes"))

                # resources
                RoleGroups.append(If(condname, m, Ref("AWS::NoValue")))

                try:
                    policy_arns = cfg.IAMGroup[m]["ManagedPolicyArns"]
                except Exception:
                    pass
                else:
                    for p in policy_arns:
                        ManagedPolicyArns.append(
                            If(
                                condname,
                                ImportValue(f"IAMPolicy{p}"),
                                Ref("AWS::NoValue"),
                            )
                        )

        # resources
        r_Role = iam.Role(f"IAMRole{n}", ManagedPolicyArns=ManagedPolicyArns)
        auto_get_props(
            r_Role,
            mapname="IAMRoleUser",
            linked_obj_name=resname,
            linked_obj_index=v["UserName"],
        )

        r_User = iam.User(resname, Groups=RoleGroups)
        auto_get_props(r_User, indexname=n, remapname=v["UserName"])

        add_obj([r_User, r_Role])


def IAM_UserToGroupAdditions(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"  # Ex. IAMUserToGroupAdditionBase

        Users = []
        for m, w in v["User"].items():
            condname = f"{resname}User{m}"
            # conditions
            add_obj(get_condition(condname, "not_equals", "none"))

            Users.append(
                If(
                    condname,
                    # for user defined in the same yaml file
                    # Ref(f'IAMUser{m}') if m in cfg.IAMUser
                    # else get_endvalue(condname),
                    get_endvalue(condname),
                    Ref("AWS::NoValue"),
                )
            )

        # resources
        r_GroupAdd = iam.UserToGroupAddition(resname, GroupName=n, Users=Users)

        add_obj([r_GroupAdd])


def IAM_Groups(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"  # Ex. IAMGroupBase

        # conditions
        add_obj(get_condition(resname, "equals", "yes", f"{resname}Enabled"))

        ManagedPolicyArns = []
        for m in v["ManagedPolicyArns"]:
            if m.startswith("arn"):
                ManagedPolicyArns.append(m)
            elif m.startswith("Ref("):
                ManagedPolicyArns.append(eval(m))
            else:
                ManagedPolicyArns.append(ImportValue(f"IAMPolicy{m}"))

        # resources
        r_Group = iam.Group(
            resname,
            GroupName=n,
            Condition=resname,
            ManagedPolicyArns=ManagedPolicyArns,
        )

        add_obj([r_Group])



import troposphere.autoscaling as asg
import troposphere.ec2 as ec2
import troposphere.cloudformation as cfm
import troposphere.policies as pol
import troposphere.applicationautoscaling as aas

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    import_user_data,
)


def AS_LaunchTemplate():
    Tags_List = []

    # Resources
    R_LaunchTemplate = ec2.LaunchTemplate(
        "LaunchTemplate",
        LaunchTemplateName=Sub("${AWS::StackName}-${EnvRole}"),
    )

    LaunchTemplateData = ec2.LaunchTemplateData("LaunchTemplateData")
    auto_get_props(
        LaunchTemplateData,
        "LaunchTemplateData",
        res_obj_type="AWS::EC2::LaunchTemplate",
    )

    if getattr(cfg, "IBOX_LAUNCH_TEMPLATE_NO_SG_EXTRA", False):
        SecurityGroups = []
    else:
        SecurityGroups = cfg.SecurityGroupsImport
    LaunchTemplateData.NetworkInterfaces[0].Groups.extend(SecurityGroups)

    R_LaunchTemplate.LaunchTemplateData = LaunchTemplateData

    ud_envrole = getattr(cfg, "IBOX_ROLE_EX", getattr(cfg, "envrole"))
    # try to get a complete user-data conf in package or in Ext dir
    user_data = import_user_data(ud_envrole)

    if not user_data:
        # if not found build a dynamic one
        for user_data_section in [
            "INIT",
            "PACKAGE",
            "SETUP",
            "APPS",
            "SERVICE",
            "ELBCHECK",
            ud_envrole,
            "END",
        ]:
            user_data.extend(import_user_data(f"SCRIPTS/{user_data_section}"))

    user_data = Join("", user_data)
    LaunchTemplateData.UserData = Base64(user_data)

    if getattr(cfg, "BottleRocket", False):
        # use if condition with both bottlerocket custom and standard cfg
        LaunchTemplateData.UserData = Base64(
            If(
                "BottleRocket",
                Join(
                    "\n",
                    [
                        get_endvalue(f"BottleRocketUserData{n}Line")
                        for n in cfg.BottleRocketUserData
                    ],
                ),
                user_data,
            ),
        )

    add_obj(R_LaunchTemplate)

    Tags = asg.Tags()
    Tags.tags = Tags_List
    return Tags


def AS_Autoscaling(key):
    LoadBalancers = []
    TargetGroups = []
    if cfg.LoadBalancerType == "Classic":
        for n in cfg.LoadBalancer:
            LoadBalancers.append(Ref(f"LoadBalancerClassic{n}"))
            setattr(
                cfg,
                "EC2InstanceUserDataELBClassicCheckLoadBalancerName",
                Ref(f"LoadBalancerClassic{n}"),
            )
    if cfg.LoadBalancerType in ["Application", "Network"]:
        for n, v in getattr(cfg, "ElasticLoadBalancingV2TargetGroup", {}).items():
            # check for enabled only for EC2 ones
            if n.startswith("EC2") and v.get("IBOX_ENABLED", True):
                TargetGroups.append(
                    Ref(f"ElasticLoadBalancingV2TargetGroupEC2LoadBalancer{n}")
                )
                setattr(
                    cfg,
                    "EC2InstanceUserDataELBV2CheckTargetGroupArn",
                    Ref(f"ElasticLoadBalancingV2TargetGroupTargetGroupLoadBalancer{n}"),
                )

    # Resources
    LaunchTemplateTags = AS_LaunchTemplate()

    R_ASG = asg.AutoScalingGroup(
        "AutoScalingGroupBase",
        LoadBalancerNames=LoadBalancers,
    )

    auto_get_props(R_ASG)

    try:
        R_ASG.TargetGroupARNs
    except Exception:
        R_ASG.TargetGroupARNs = TargetGroups
    else:
        R_ASG.TargetGroupARNs.extend(TargetGroups)

    R_ASG.Tags += LaunchTemplateTags

    add_obj([R_ASG])



import yaml
import yaml.constructor
import sys
import os
import copy
import json
import logging
from pprint import pprint, pformat

from . import cfg


def RP_to_cfg(key, prefix="", overwrite=True, mappedvalues=[]):
    if hasattr(key, "items"):
        for k, v in key.items():
            # remove both * and + that can be present for special IBOX usage
            key_name = f"{prefix}{k}".translate("".maketrans({"*": None, "+": None}))
            try:
                getattr(cfg, key_name)
                exist = True
            except Exception:
                exist = False
            if not exist or overwrite:
                setattr(cfg, key_name, v)
                if key_name not in mappedvalues:
                    cfg.fixedvalues[key_name] = v
            # recursively traverse dict
            # keys name are the concatenation of traversed dict keys
            if isinstance(v, dict):
                for j, w in v.items():
                    RP_to_cfg({f"{k}{j}": w}, prefix, overwrite)


def merge_dict(base, work):
    if isinstance(work, (str, list)) or not base:
        return work
    keys = dict(list(base.items()) + list(work.items())).keys()
    for k in keys:
        if k.endswith("**"):
            # ** is used to replace existing dict instead of merging it
            base[k.replace("**", "")] = work[k]
        elif isinstance(base.get(k), dict) and isinstance(work.get(k), dict):
            base[k] = merge_dict(base[k], work[k])
        elif k.endswith("++") and isinstance(work.get(k), list):
            # ++ is used to append elements to an existing key
            try:
                base[k.replace("++", "")] += work[k]
            except Exception:
                base[k.replace("++", "")] = work[k]
        elif k in work:
            base[k] = work[k]
    return base


def build_RP():
    LD_INCLUDED = []
    LD_EXCLUDED = []

    CFG_FILE_INT = "%s/cfg" % os.path.dirname(os.path.realpath(__file__))
    CFG_FILE_INT = os.path.normpath(CFG_FILE_INT)

    CFG_FILE_EXT = "%s/cfg" % os.getcwd()
    CFG_FILE_EXT = os.path.normpath(CFG_FILE_EXT)

    envrole = cfg.envrole
    stacktype = cfg.stacktype
    brand = cfg.brand

    RP_base = {}

    # dynamically populate RP_base Dict from ENV_BASE and cfg.regions
    for n in cfg.ENV_BASE:
        RP_base[n] = {}
        for m in cfg.regions:
            RP_base[n][m] = {}

    cfg.RP_base = RP_base

    def _get_RP_base_keys():
        RP_base_keys = ["global"]
        for n, v in RP_base.items():
            RP_base_keys.append(n)
            for m, w in v.items():
                RP_base_keys.append(m)

        return RP_base_keys + [stacktype, envrole]

    RP_base_keys = _get_RP_base_keys()

    class Loader(yaml.Loader):
        def __init__(self, stream):
            # This way for include relative to file with include statement
            self._root_current = os.path.split(stream.name)[0]
            # This way for include BASE relative on BASE dir
            self._root_base = os.path.join(CFG_FILE_INT, "BASE")
            # This way for include relative on BASE EXT dir
            self._root_base_ext = os.path.join(CFG_FILE_EXT, "BASE")
            # This way for include relative on brand EXT dir
            self._root_brand_ext = os.path.join(CFG_FILE_EXT, brand)

            # try to find out if source file is in a subfolder
            if self._root_base in self._root_current:
                suffix = self._root_current.replace(self._root_base, "")
            elif self._root_base_ext in self._root_current:
                suffix = self._root_current.replace(self._root_base_ext, "")
            else:
                suffix = self._root_current.replace(self._root_brand_ext, "")

            # self.root_current_suffix will be empty if source file is in root
            self.root_current_suffix = os.path.basename(suffix)
            if self.root_current_suffix.upper() in cfg.STACK_TYPES:
                # set to empty even if UPPERCASE subfolders is in
                # cfg.STACK_TYPES to not be searched for include files
                self.root_current_suffix = ""

            self.stream = stream
            super(Loader, self).__init__(stream)
            Loader.add_constructor("!include", Loader.include)
            Loader.add_constructor("!exclude", Loader.exclude)

        def exclude(self, node):
            if isinstance(node, yaml.SequenceNode):
                for filename in self.construct_sequence(node):
                    LD_EXCLUDED.append(filename)

        def include(self, node):
            if isinstance(node, yaml.SequenceNode):
                result = []
                for filename in self.construct_sequence(node):
                    if filename in LD_EXCLUDED + LD_INCLUDED:
                        # a yml file can be included only once
                        continue
                    LD_INCLUDED.append(filename)

                    # search for include in files in roots
                    for path in [
                        self._root_base,
                        self._root_base_ext,
                        self._root_brand_ext,
                    ]:
                        if "/" not in filename and self.root_current_suffix:
                            # for file in subfolder that include file without a path
                            # search only in current subfolder
                            path = os.path.join(path, self.root_current_suffix)

                        contents = self.extractFile(filename, path)
                        if contents:
                            result.append(contents)
                return result
            else:
                print("Error:: unrecognised node type in !include statement")
                raise yaml.constructor.ConstructorError

        def extractFile(self, filename, root):
            filepath = os.path.join(root, filename)
            try:
                with open(filepath, "r") as f:
                    return yaml.load(f, Loader)
            except IOError:
                pass

    def process_cfg(cfg, envs, skip=True):
        if hasattr(cfg, "items"):
            # This method allow to delete items from a dictionary
            # while iterating over it
            for k in list(cfg):
                v = cfg[k]
                # for final values
                if isinstance(v, (str, int, list)) and not k.startswith("IBoxLoader"):
                    yield {k: v}
                # for empty dict in common.yml
                if isinstance(v, dict) and len(v) == 0:
                    yield {k: v}
                # for recursively descending in env/region role dict.
                # list is needed for IBoxLoader include list.
                if k in envs and isinstance(v, (dict, list)):
                    if k in ["IBoxLoader", "IBoxLoaderAfter"]:
                        # IBoxLoader included dict processed by
                        # process_cfg keep skip = True
                        skip = True
                    else:
                        skip = False
                    try:
                        # after descending in env main key
                        # (not the one nested under region) delete key
                        # this way when envs include both (env and region)
                        # i do not process it again
                        if k in list(RP_base.keys()):
                            del cfg[k]
                    except Exception:
                        pass
                    for result in process_cfg(v, envs, skip=skip):
                        yield result
                if skip:
                    # Here i do not have yet encountered a envs key
                    # skip properties until i find it.
                    # Useful for using yaml anchors before global key
                    continue
                # for recursively descending in dict not in RP_base_keys
                # (env/region/envrole/stacktype)
                # (final key is the concatenation of traversed dict keys)
                if k not in RP_base_keys and isinstance(v, dict):
                    for j, w in v.items():
                        for result in process_cfg({f"{k}{j}": w}, envs, skip=skip):
                            yield result
        if isinstance(cfg, list):
            for n in cfg:
                for result in process_cfg(n, envs, skip=skip):
                    yield result

    def replace_not_allowed_char(s):
        # CF Mapping allow for index only alfanumeric char
        # this way i can specify more "clear" name
        # for index in CloudFormation behaviours
        key = str(s)
        for s, w in cfg.CLF_PATH_PATTERN_REPLACEMENT.items():
            key = key.replace(s, w)

        return int(key) if key.isdigit() else key

    def merge_RP(data):
        def _process(key, data, RP, merge=True):
            key = str(key)

            if key.startswith("/"):
                # for CFront Behaviors
                key = replace_not_allowed_char(key)
            if key.endswith("**"):
                # ** is used to replace existing dict instead of merging it
                f_key = key.replace("**", "")
                if isinstance(RP.get(f_key), dict):
                    # replace key only if same key already exist and is a dict
                    key = f_key
                merge = False

            if merge and isinstance(RP.get(key), dict):
                # RP[key] already exist as a dict, try merging
                RP[key] = merge_dict(RP[key], _recurse(data))
            elif isinstance(RP.get(key), str) and RP[key] == "IBOX_SKIP_FUNC":
                # if key need to be skipped using IBOX_SKIP_FUNC avoid overwriting it
                pass
            elif key in cfg.MERGE_RP_KEEP_AS_LIST and isinstance(data, list):
                # this keys need to stay as a list of dicts
                RP[key] = data
            else:
                RP[key] = _recurse(data)

        def _recurse(data):
            RP = {}
            if isinstance(data, dict):
                # data is a dict
                for n, v in data.items():
                    _process(n, v, RP)
            elif isinstance(data, list) and data and isinstance(data[0], dict):
                # data is a list of dicts
                for v in data:
                    for m, w in v.items():
                        _process(m, w, RP)
            else:
                RP = data
            return RP

        return _recurse(data)

    def read_yaml(file_type, brand, base_dir, stacktype=""):
        cfg_file = os.path.join(base_dir, brand, stacktype.upper(), f"{file_type}.yml")

        try:
            with open(cfg_file, "r") as ymlfile:
                cfg = yaml.load(ymlfile, Loader=Loader)

                return cfg
        except IOError:
            return {}

    def inject_ibox_base(RP, root=""):
        base_key = "IBOX_BASE"
        for main_key in list(RP.keys()):
            main_key_full = f"{root}{main_key}"
            main_key_value = RP[main_key]
            if isinstance(main_key_value, dict) and base_key in main_key_value:
                base_key_value = main_key_value[base_key]
                base_key_value["IBOX_BASE_REF"] = True

                for resource_key, resource_key_value in main_key_value.items():
                    if (
                        resource_key == base_key
                        or "IBOX_BASE_SKIP" in resource_key_value
                    ):
                        continue
                    # inject in cfg key/value
                    RP_to_cfg(
                        base_key_value,
                        prefix=f"{main_key_full}{resource_key}",
                        overwrite=False,
                    )

                    # inject in existing structure
                    merged = merge_dict(
                        copy.deepcopy(base_key_value), resource_key_value
                    )
                    # need to do it this way to keep the "link" between existing dict keys and values
                    for n, v in merged.items():
                        RP[main_key][resource_key][n] = v
                del RP[main_key][base_key]

            if isinstance(main_key_value, dict):
                inject_ibox_base(main_key_value, root=main_key_full)

    def get_RP(yaml_cfg):
        cfg_keys = ["IBoxLoader", "IBoxLoaderAfter"]

        def _parse_yaml(conf, envs=[]):
            parsed_yaml = {}
            for value in process_cfg(conf, envs):
                for k, v in value.items():
                    try:
                        # if v is a list of dict and the same key (k) already
                        # exist, final value is the sum of the list (of dicts)
                        if isinstance(v[0], dict):
                            parsed_yaml[k] = parsed_yaml[k] + v
                        else:
                            raise
                    except Exception:
                        parsed_yaml[k] = v
            return parsed_yaml

        def get_RP_tree():
            RP_list = []

            for cfg_type, ctv in yaml_cfg.items():
                for read_yaml in ctv:
                    parsed_global = _parse_yaml(read_yaml, cfg_keys + ["global"])
                    RP_list.append(parsed_global)

            # merge keys value
            RP = merge_RP(RP_list)

            return RP

        def get_RP_map():
            RP = copy.deepcopy(RP_base)
            mapped_keys = []

            # I build the cfg to build the mappings for env/region
            for env in list(RP):
                rvalue = RP[env]
                env_cfg = {}
                # first the configuration under the env key
                for cfg_type, ctv in yaml_cfg.items():
                    for read_yaml in ctv:
                        env_cfg.update(_parse_yaml(read_yaml, cfg_keys + [env]))
                # then the one under region + env keys
                for region in list(rvalue.keys()):
                    RP[env][region] = copy.deepcopy(env_cfg)
                    for cfg_type, ctv in yaml_cfg.items():
                        for read_yaml in ctv:
                            RP[env][region].update(
                                _parse_yaml(
                                    copy.deepcopy(read_yaml), cfg_keys + [env, region]
                                )
                            )
                    # create list of all mapped keys
                    for key in list(RP[env][region]):
                        if hasattr(cfg, f"{key}@"):
                            # if there is key in RP_tree/cfg and it endswith "@"
                            # skip mapped values and update cfg.fixedvalues and RP
                            cfg.fixedvalues[key] = getattr(cfg, f"{key}@")
                            del RP[env][region][key]
                        elif hasattr(cfg, key) and RP[env][region][key] == getattr(
                            cfg, key
                        ):
                            # if value is equal to the global one, delete the key
                            del RP[env][region][key]
                        elif key not in mapped_keys:
                            # only add if not already present
                            mapped_keys.append(key)
                            # delete the fixed one
                            try:
                                del cfg.fixedvalues[key]
                            except Exception:
                                pass

            cfg.mappedvalues = mapped_keys
            return RP

        # RP_tree represent the resources structure and it's configuration.
        RP_tree = get_RP_tree()

        # Read RP_tree and put a flat key/value configuration in cfg.
        cfg.fixedvalues = {}
        RP_to_cfg(RP_tree)

        # Inject IBOX_BASE configurations
        inject_ibox_base(RP_tree)

        # Create the mapping for env/region.
        RP_map = get_RP_map()

        return RP_tree, RP_map

    # End inner methods and begin of main code.

    # envrole type files must be read first
    yaml_role = [
        read_yaml(envrole, "BASE", CFG_FILE_INT, stacktype),
        read_yaml(envrole, "BASE", CFG_FILE_EXT, stacktype),
        read_yaml(envrole, brand, CFG_FILE_EXT, stacktype),
    ]

    yaml_cfg = {
        "common": [
            read_yaml("common", "BASE", CFG_FILE_INT),
            read_yaml("common", "BASE", CFG_FILE_EXT),
            read_yaml("common", brand, CFG_FILE_EXT),
        ],
        "type": [
            read_yaml("TYPE", "BASE", CFG_FILE_INT, stacktype),
            read_yaml("TYPE", "BASE", CFG_FILE_EXT, stacktype),
            read_yaml("TYPE", brand, CFG_FILE_EXT, stacktype),
        ],
        "role": yaml_role,
    }

    cfg.RP_tree, cfg.RP_map = get_RP(yaml_cfg)

    if cfg.debug:
        print("########## RP ######### START #####")
        pprint(cfg.RP_tree)
        print("########## RP ######### END #######")

        print("########## EXCLUDED ####### START #####")
        pprint(LD_EXCLUDED)
        print("########## EXCLUDED ####### END #######")

        print("########## INCLUDED ####### START #####")
        pprint(LD_INCLUDED)
        print("########## INCLUDED ####### END #######")

        print("########## FIXEDVALUES ######### START #####")
        pprint(cfg.fixedvalues)
        print("########## FIXEDVALUES ######### END #######")

        print("########## MAPPEDVALUES ######### START #####")
        pprint(cfg.RP_map)
        print("########## MAPPEDVALUES ######### END #######")



import troposphere.cloudfront as clf

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    change_obj_data,
    clf_compute_order,
)


def process_cache_policy(v):
    # Use CachePolicyId/OriginRequestPolicyId or legacy mode
    if "CachePolicyId" in v:
        for k in ["DefaultTTL", "MaxTTL", "MinTTL", "ForwardedValues"]:
            try:
                del v[k]
            except Exception:
                pass


def cache_behavior_process(key):
    # process default behavior
    process_cache_policy(
        key["DefaultCacheBehavior"]
    )
    # process other behaviors
    for n, v in key["CacheBehaviors"].items():
        process_cache_policy(v)


def origin_process(name, key):
    for n, v in key["Origins"].items():
        resname = f"{name}DistributionConfigOrigins{n}"

        if "OriginAccessIdentity" in v.get("S3OriginConfig", []):
            del v["CustomOriginConfig"]
        else:
            del v["S3OriginConfig"]


def CF_CloudFront(key):
    for n, v in getattr(cfg, key).items():
        if not v.get("IBOX_ENABLED", True):
            continue
        resname = f"{key}{n}"
        # Resources
        R_CloudFrontDistribution = clf.Distribution(resname)
        distribution_config = v["DistributionConfig"]

        # process cache_behavior and origin
        cache_behavior_process(distribution_config)
        origin_process(resname, distribution_config)

        # Automatically compute Behaviour Order based on PathPattern
        cfg.dbg_clf_compute_order = {}
        sortedcachebehaviors = sorted(
            distribution_config["CacheBehaviors"].items(),
            key=lambda x_y: clf_compute_order(x_y[1]["PathPattern"]),
        )
        distribution_config["CacheBehaviors"] = {x[0]: x[1] for x in sortedcachebehaviors}

        if cfg.debug:
            print("##########CLF_COMPUTE_ORDER#########START#######")
            for n, v in iter(
                sorted(cfg.dbg_clf_compute_order.items(), key=lambda x_y: x_y[1])
            ):
                print(f"{n} {v}\n")
            print("##########CLF_COMPUTE_ORDER#########END#######")

        auto_get_props(R_CloudFrontDistribution, mapname=resname)
        add_obj(R_CloudFrontDistribution)



import troposphere.s3 as s3

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    get_dictvalue,
)


def S3_Buckets(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"
        bucket_name = getattr(cfg, f"BucketName{n}")

        # resources
        r_Bucket = s3.Bucket(resname)
        auto_get_props(r_Bucket, remapname=bucket_name, indexname=n)

        r_Policy = s3.BucketPolicy(
            f"BucketPolicy{n}",
            Condition=resname,
            Bucket=Ref(r_Bucket.title),
        )
        # BucketPolicy Statements are read from yaml cfg, so update it with dynamic data
        base_statements = cfg.S3BucketPolicyBasePolicyDocumentStatement
        # At least one statement must be always present, create a simple one with no conditions
        base_statements["AllowReplica"]["Resource"] = getattr(
            cfg, f"{resname}PolicyStatementReplicaResourcePrefix"
        )
        base_statements["AllowAccountsRead"]["Principal"]["AWS"] = getattr(
            cfg, f"{resname}PolicyStatementAccountsReadPrincipal"
        )
        base_statements["AllowAccountsWrite"]["Principal"]["AWS"] = getattr(
            cfg, f"{resname}PolicyStatementAccountsWritePrincipal"
        )
        base_statements["AllowAccountsDelete"]["Principal"]["AWS"] = getattr(
            cfg, f"{resname}PolicyStatementAccountsDeletePrincipal"
        )
        auto_get_props(
            r_Policy,
            mapname="S3BucketPolicyBase",
            linked_obj_name=bucket_name,
            linked_obj_index=resname,
        )

        BucketPolicyStatement = r_Policy.PolicyDocument["Statement"]

        # BucketPolicy key
        if "BucketPolicy" in v:
            for _, bp in v["BucketPolicy"].items():
                BucketPolicyStatement.append(get_dictvalue(bp))

        # Policy CloudFrontOriginAccessIdentity
        PolicyStatementCloudFrontOriginAccessIdentityPrincipal = getattr(
            cfg,
            f"{resname}PolicyStatementCloudFrontOriginAccessIdentityPrincipal",
            None,
        )
        if isinstance(PolicyStatementCloudFrontOriginAccessIdentityPrincipal, list):
            # resources
            BucketPolicyStatement.append(
                {
                    "Action": ["s3:GetObject"],
                    "Effect": "Allow",
                    "Resource": [Sub("arn:aws:s3:::%s/*" % bucket_name)],
                    "Principal": {
                        "AWS": PolicyStatementCloudFrontOriginAccessIdentityPrincipal
                    },
                    "Sid": "AllowCFAccess",
                }
            )

        add_obj([r_Bucket, r_Policy])



import troposphere.waf as waf
import troposphere.wafregional as wafr

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


class WAFIPSet(waf.IPSet):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title  # Ex. WafIPSetAwsNat
        self.Name = name  # Ex. AwsNat


class WAFByteMatchSet(waf.ByteMatchSet):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title
        self.Name = name


class WAFRule(waf.Rule):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title  # Ex. WafRuleBlockIPsNotAllowed
        self.Name = name  # Ex. BlockIPsNotAllowed
        self.MetricName = self.Name


class WAFWebACL(waf.WebACL):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title
        self.Name = name
        self.MetricName = self.Name


class WAFGlobalIPSet(WAFIPSet):
    pass


class WAFGlobalByteMatchSet(WAFByteMatchSet):
    pass


class WAFGlobalRule(WAFRule):
    pass


class WAFGlobalWebAcl(WAFWebACL):
    pass


class WAFRegionalIPSet(wafr.IPSet, WAFIPSet):
    pass


class WAFRegionalByteMatchSet(wafr.ByteMatchSet, WAFByteMatchSet):
    pass


class WAFRegionalRule(wafr.Rule, WAFRule):
    pass


class WAFRegionalWebAcl(wafr.WebACL, WAFWebACL):
    pass


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class WAFIPSetDescriptors(waf.IPSetDescriptors, wafr.IPSetDescriptors):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Type = "IPV4"
        self.Value = Select(int(self.title), get_endvalue(f"{name}Ips"))


class WAFPredicates(waf.Predicates, wafr.Predicates):
    def __init__(self, title, name, ptype, wtype, key, **kwargs):
        super().__init__(title, **kwargs)
        self.Negated = get_endvalue(f"{name}Negated")

        try:
            self.Type = key["Type"]
        except Exception:
            pass
        else:
            ptype = self.Type

        if ptype == "ByteMatch":
            self.DataId = Ref(f"Waf{wtype}ByteMatchSet{self.title}")
            self.Type = "ByteMatch"
        if ptype == "IPMatch":
            self.DataId = Ref(f"Waf{wtype}IPSet{self.title}")
            self.Type = "IPMatch"
        if ptype == "SizeConstraint":
            self.DataId = Ref(f"Waf{wtype}SizeConstraintSet{self.title}")
            self.Type = "SizeConstraint"
        if ptype == "SqlInjectionMatch":
            self.DataId = Ref(f"Waf{wtype}SqlInjectionMatchSet{self.title}")
            self.Type = "SqlInjectionMatch"
        if ptype == "XssMatch":
            self.DataId = Ref(f"Waf{wtype}XssMatchSet{self.title}")
            self.Type = "XssMatch"


class WAFAction(waf.Action, wafr.Action):
    pass


class WAFWebACLRule(waf.Rules, wafr.Rules):
    def __init__(self, title, name, index, wtype, **kwargs):
        super().__init__(title, **kwargs)
        self.Action = WAFAction(Type=get_endvalue(f"{name}Action"))
        self.Priority = index
        self.RuleId = Ref(f"Waf{wtype}Rule{self.title}")


def WAF_condition(condname, mapname, wtype):
    return {
        condname: And(
            get_condition("", "equals", "yes", f"{mapname}Enabled"),
            Condition("Global") if wtype == "Global" else Equals("1", "1"),
            Or(
                get_condition("", "equals", wtype, f"{mapname}WafType"),
                get_condition("", "equals", "Common", f"{mapname}WafType"),
            ),
        )
    }


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def WAF_IPSets(key, wtype=""):
    for n, v in getattr(cfg, key).items():
        name = key.replace("Waf", wtype)  # Ex. RegionalIPSet
        resname = f"Waf{name}{n}"  # Ex. WafRegionalIPSetAwsNat
        mapname = f"{key}{n}"  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        IPSetDescriptors = []
        for i, j in enumerate(v["Ips"]):
            IPSetDescriptor = WAFIPSetDescriptors(str(i), name=mapname)
            IPSetDescriptors.append(IPSetDescriptor)

        IPSet = globals()[f"WAF{name}"](resname, name=n)
        IPSet.IPSetDescriptors = IPSetDescriptors

        add_obj(IPSet)


def WAF_ByteMatchSets(key, wtype=""):
    for n, v in getattr(cfg, key).items():
        name = key.replace("Waf", wtype)  # Ex. RegionalByteMatchSet
        resname = f"Waf{name}{n}"
        mapname = f"{key}{n}"

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        r_ByteMatchSet = globals()[f"WAF{name}"](resname, name=n)
        auto_get_props(r_ByteMatchSet, mapname)

        add_obj(r_ByteMatchSet)


def WAF_Rules(key, wtype=""):
    for n, v in getattr(cfg, key).items():
        name = key.replace("Waf", wtype)  # Ex. RegionalIPSet
        resname = f"Waf{name}{n}"  # Ex. WafRegionalIPSetAwsNat
        mapname = f"{key}{n}"  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        Predicates = []
        O_Predicates = []
        for m, w in v["Predicates"].items():
            predmapname = f"{mapname}Predicates{m}"
            predresname = m
            ptype = v["Type"]
            Predicate = WAFPredicates(
                predresname, name=predmapname, ptype=ptype, wtype=wtype, key=w
            )

            Predicates.append(
                If(Predicate.DataId.data["Ref"], Predicate, Ref("AWS::NoValue"))
            )

            O_Predicates.append(
                If(Predicate.DataId.data["Ref"], m, Ref("AWS::NoValue"))
            )

        Rule = globals()[f"WAF{name}"](resname, name=n)
        Rule.Predicates = Predicates

        add_obj(Rule)

        # outputs
        O_Rule = Output(f"{mapname}{wtype}")
        O_Rule.Condition = resname
        O_Rule.Value = Join(",", O_Predicates)

        add_obj(O_Rule)


def WAF_WebAcls(key, wtype=""):
    for n, v in getattr(cfg, key).items():
        name = key.replace("Waf", wtype)  # Ex. RegionalIPSet
        resname = f"Waf{name}{n}"  # Ex. WafRegionalIPSetAwsNat
        mapname = f"{key}{n}"  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        Rules = []
        O_Rules = []
        for w, m in enumerate(v["Rules"], start=1):
            rulemapname = f"{mapname}Rules{m}"
            ruleresname = m  # Ex. BlockIPsNotAllowed
            Rule = WAFWebACLRule(ruleresname, name=rulemapname, index=w, wtype=wtype)

            Rules.append(Rule)

            O_Rules.append(m)

        WebACL = globals()[f"WAF{name}"](resname, name=n)
        WebACL.Rules = Rules
        WebACL.DefaultAction = WAFAction(Type=v["DefaultAction"])

        add_obj(WebACL)

        # outputs
        O_WebACL = Output(f"{mapname}{wtype}")
        O_WebACL.Condition = resname
        O_WebACL.Value = Sub("${%s} - %s" % (resname, ",".join(O_Rules)))

        add_obj(O_WebACL)


def WAF_GlobalByteMatchSets(key):
    WAF_ByteMatchSets(key, wtype="Global")


def WAF_GlobalIPSets(key):
    WAF_IPSets(key, wtype="Global")


def WAF_GlobalRules(key):
    WAF_Rules(key, wtype="Global")


def WAF_GlobalWebAcls(key):
    WAF_WebAcls(key, wtype="Global")


def WAF_RegionalByteMatchSets(key):
    WAF_ByteMatchSets(key, wtype="Regional")


def WAF_RegionalIPSets(key):
    WAF_IPSets(key, wtype="Regional")


def WAF_RegionalRules(key):
    WAF_Rules(key, wtype="Regional")


def WAF_RegionalWebAcls(key):
    WAF_WebAcls(key, wtype="Regional")



import troposphere.efs as efs

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)
from .ec2 import SecurityGroup, SecurityGroupIngress


class EFSMountTarget(efs.MountTarget):
    def __init__(self, title, index, sgname, efsresname, **kwargs):
        super().__init__(title, **kwargs)
        name = self.title
        self.Condition = name
        self.FileSystemId = Ref(efsresname)
        self.SecurityGroups = [Ref(sgname)]
        self.SubnetId = Select(str(index), Split(",", get_expvalue("SubnetsPrivate")))


# #################################
# ### START STACK INFRA CLASSES ###
# #################################


def EFS_FileStorage(key):
    for n, v in getattr(cfg, key).items():
        resname = f"{key}{n}"  # Ex. EFSFileSystemWordPress
        sgservername = f"SecurityGroupEFSServer{n}"
        sgclientname = f"SecurityGroupEFS{n}"
        sginame = f"SecurityGroupIngressEFS{n}"
        for i in range(cfg.AZones["MAX"]):
            mountname = f"EFSMountTarget{n}{i}"
            # conditions
            add_obj(
                {
                    mountname: And(
                        Condition(resname),
                        Equals(
                            FindInMap("AvabilityZones", Ref("AWS::Region"), f"Zone{i}"),
                            "True",
                        ),
                    )
                }
            )

            # resources
            r_Mount = EFSMountTarget(
                mountname, index=i, sgname=sgservername, efsresname=resname
            )

            add_obj(r_Mount)
        # conditions
        add_obj(get_condition(resname, "equals", "yes", f"{resname}Enabled"))

        # resources
        r_File = efs.FileSystem(resname)
        auto_get_props(r_File, indexname=n)

        r_SGServer = SecurityGroup(
            sgservername,
            Condition=resname,
            GroupDescription=f"Rule to access EFS FileSystem {n}",
        )

        r_SGClient = SecurityGroup(
            sgclientname,
            Condition=resname,
            GroupDescription=f"Enable access to EFS FileSystem {n}",
        )

        SGIExtra = {
            "Name": sginame,
            "Condition": resname,
            "FromPort": 2049,
            "GroupId": "Sub('${%s.GroupId}')" % sgservername,
            "SourceSecurityGroupId": "Ref('%s')" % sgclientname,
            "ToPort": 2049,
        }

        r_SGI = SecurityGroupIngress(sginame)
        auto_get_props(r_SGI, rootdict=SGIExtra)

        # outputs
        o_SGClient = Output(
            sgclientname,
            Condition=resname,
            Value=GetAtt(sgclientname, "GroupId"),
            Export=Export(sgclientname),
        )

        add_obj([r_File, r_SGServer, r_SGClient, r_SGI, o_SGClient])



import troposphere.ecs as ecs

from .common import *
from .shared import (
    Parameter,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
)


def ECS_TaskDefinition(key):
    for n, v in getattr(cfg, key).items():
        if not v.get("IBOX_ENABLED", True):
            continue
        mapname = f"{key}{n}"

        # Resources
        R_TaskDefinition = ecs.TaskDefinition(mapname)
        auto_get_props(R_TaskDefinition)

        R_TaskDefinition.ContainerDefinitions = []
        for n, v in cfg.ContainerDefinitions.items():
            name = f"ContainerDefinitions{n}"  # Ex. ContainerDefinitions1
            r_Container = ecs.ContainerDefinition(name)
            auto_get_props(r_Container, indexname=n, res_obj_type="AWS::ECS::TaskDefinition")
            R_TaskDefinition.ContainerDefinitions.append(r_Container)

        add_obj([R_TaskDefinition])



#!/usr/bin/env python3
import sys
import os
import json
import concurrent.futures
from traceback import print_exc

from troposphere import Template
from awsibox import args, cfg, RP, discover, generate

args = args.get_args()


def output_template(template, brand=None, envrole=None):
    if args.Output == "json":
        output = template.to_json()
        extension = ".json"
    elif args.Output == "cjson":
        output = template.to_json(indent=None)
        extension = ".json"
    elif args.Output == "yaml":
        output = template.to_yaml(long_form=True)
        extension = ".yaml"

    if args.action == "view":
        print(output)
    elif args.action == "write":
        file_path = os.path.join(os.getcwd(), "templates", brand)
        try:
            os.makedirs(file_path)
        except Exception:
            pass
        file_name = os.path.join(file_path, envrole + extension)
        with open(file_name, "w") as f:
            f.write(output + "\n")

        print(f"Brand: {brand} - EnvRole: {envrole}")


def get_template():
    # save base cfg
    cfg_base = cfg.__dict__.copy()

    cfg.template = Template()
    RP.build_RP()

    try:
        template = generate.generate()
    finally:
        # restore base cfg (every dynamic dict in cfg need to be cleared)
        cfg.OBJS.clear()
        cfg.__dict__.clear()
        cfg.__dict__.update(cfg_base)

    return template


def do_process(stacktype, envrole):
    cfg.envrole = envrole
    cfg.stacktype = stacktype
    template = get_template()
    output_template(template, cfg.brand, envrole)


def concurrent_exec(roles, kwargs):
    with concurrent.futures.ProcessPoolExecutor(**kwargs) as executor:
        data = {}
        future_to_role = {}
        for n in roles:
            stacktype = n[0]
            envrole = n[1]
            ex_sub = executor.submit(do_process, stacktype, envrole)
            future_to_role[ex_sub] = envrole

        for future in concurrent.futures.as_completed(future_to_role):
            role = future_to_role[future]
            try:
                data[role] = future.result()
            except Exception as e:
                print(f"{role} generated an exception: {e}")
                print_exc()
                break
        for future in future_to_role:
            future.cancel()

# read CloudFormationResourceSpecification to get CloudFormation Resources Properties
with open(f"{cfg.PATH_INT}/BASE/CloudFormationResourceSpecification.json", "r") as cfm_res_spec_json:
    cfg.cfm_res_spec = json.load(cfm_res_spec_json)

if args.action == "view":
    discover_map = discover.discover([args.Brand], [args.EnvRole], [])

    cfg.brand = args.Brand
    try:
        role = discover_map[cfg.brand][0]
    except Exception:
        pass
    else:
        do_process(role[0], role[1])
elif args.action == "write":
    discover_map = discover.discover(args.Brands, args.EnvRoles, args.StackTypes)

    for brand, roles in discover_map.items():
        cfg.brand = brand
        if args.jobs:
            kwargs = {"max_workers": args.jobs}
        else:
            kwargs = {}
        concurrent_exec(set(roles), kwargs)



import troposphere.cloudformation as cfm

from .common import *
from .shared import (
    Parameter,
    do_no_override,
    get_endvalue,
    get_expvalue,
    get_subvalue,
    auto_get_props,
    get_condition,
    add_obj,
    auto_build_obj,
)


def CFM_Parameters(key):
    auto_build_obj(Parameter(""), getattr(cfg, key))


def CFM_Conditions(key):
    do_no_override(True)
    for n, v in getattr(cfg, key).items():
        c_Condition = {n: eval(v)}

        add_obj(c_Condition)
    do_no_override(False)


def CFM_Mappings(key):
    for n, v in getattr(cfg, key).items():
        c_Mapping = {n: v}

        cfg.Mappings.update(c_Mapping)


def CFM_Outputs(key):
    auto_build_obj(Output(""), getattr(cfg, key))


def CFM_CustomResourceReplicator(key):
    resname = "CloudFormationCustomResourceStackReplicator"
    # Parameters
    P_ReplicateRegions = Parameter(
        "CCRReplicateRegions",
        Description="Regions where to replicate - none to disable - empty for default based on env/role",
        Type="CommaDelimitedList",
    )

    add_obj(P_ReplicateRegions)

    # Resources
    R_Replicator = cfm.CustomResource(resname)

    if "LambdaCCRStackReplicator" in cfg.Resources:
        R_Replicator.DependsOn = "IAMPolicyLambdaCCRStackReplicator"
        R_Replicator.ServiceToken = GetAtt("LambdaCCRStackReplicator", "Arn")
    else:
        R_Replicator.ServiceToken = get_expvalue("LambdaCCRStackReplicator")

    for p, v in cfg.Parameters.items():
        if not p.startswith("Env"):
            value = get_endvalue(p)
        else:
            value = Ref(p)
        setattr(R_Replicator, p, value)

    add_obj(R_Replicator)


def CFM_CustomResourceLightHouse(key):
    resname = "CloudFormationCustomResourceLightHouse"
    # Parameters
    P_LightHouse = Parameter(
        "CCRLightHouse",
        Description="Enable CustomResource for LightHouse - " "empty for mapped value",
        AllowedValues=["", "yes", "no"],
    )

    add_obj(P_LightHouse)

    # Conditions
    C_LightHouse = get_condition(resname, "equals", "yes", "CCRLightHouse")

    add_obj(C_LightHouse)

    # Resources
    R_LightHouse = cfm.CustomResource(
        resname,
        Condition=resname,
        DependsOn="Service",
        ServiceToken=get_expvalue("LambdaCCRLightHouse"),
        EnvRole=Ref("EnvRole"),
        EnvApp1Version=Ref("EnvApp1Version"),
        RepoName=getattr(cfg, "ContainerDefinitions1ImageSuffix").split(":")[0],
    )

    add_obj(R_LightHouse)


def CFM_CustomResourceFargateSpot(key):
    resname = "CloudFormationCustomResourceFargateSpot"
    R_FargateSpot = cfm.CustomResource(
        resname,
        Condition="FargateSpot",
        DependsOn="ServiceSpot",
        ServiceToken=get_expvalue("LambdaCCRFargateSpot"),
        ServiceArn=Ref("ServiceSpot"),
        ServiceBase=GetAtt("Service", "Name"),
        ServiceSpot=GetAtt("ServiceSpot", "Name"),
        Cluster=get_expvalue("Cluster", "ClusterStack"),
        ScalingPolicy=Ref("ApplicationAutoScalingScalingPolicyCpu"),
    )

    add_obj(R_FargateSpot)



import setuptools

setuptools.setup(
    packages=[
        'awsibox',
    ],
    package_data={
        'awsibox': ['cfg/BASE/*'],
        'awsibox': ['templates/*'],
        'awsibox': ['lambdas/*'],
        'awsibox': ['user-data/*'],
    },
    install_requires=[
        'troposphere',
        'PyYAML>=5,==5.*',
        'python_minifier',
    ],
    python_requires='>=3.7',
    scripts=[
        'scripts/ibox_generate_templates.py',
    ],
    include_package_data=True,
)



from pprint import pprint, pformat
import argparse

from . import cfg


# parse main argumets
def get_args():
    parser = argparse.ArgumentParser(
        description="Generate AWS CloudFormation Stack Templates"
    )

    # subparser
    subparsers = parser.add_subparsers(
        help="Desired Action", required=True, dest="action"
    )

    # parent parser args
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-R",
        "--Regions",
        help="Regions to enable",
        nargs="+",
        default=cfg.DEFAULT_REGIONS,
    )
    parent_parser.add_argument(
        "-o",
        "--Output",
        help="Output Format",
        choices=["json", "cjson", "yaml"],
        default="yaml",
    )
    parent_parser.add_argument(
        "-d", "--Debug", help="Show RP Dict", action="store_true"
    )

    # view parser
    parser_view = subparsers.add_parser(
        "view",
        parents=[parent_parser],
        help="Generate template and print to standard output",
    )
    parser_view.add_argument(
        "-r", "--EnvRole", help="Stack EnvRole", required=True, type=str
    )
    parser_view.add_argument("-b", "--Brand", help="Brand", required=True, type=str)

    # write parser
    parser_write = subparsers.add_parser(
        "write",
        parents=[parent_parser],
        help=("Generate template " 'and write to "template/${brand}/$envrole" file'),
    )

    parser_write.add_argument(
        "-b",
        "--Brands",
        help="Limit generation to EnvRoles defined in the specified Brands",
        nargs="+",
    )
    parser_write.add_argument(
        "-j", "--jobs", help="Max concurrent template generation jobs", type=int
    )

    parser_write.add_argument(
        "-r",
        "--EnvRoles",
        help="Limit generation to specified EnvRoles",
        nargs="+",
        default=[],
    )

    parser_write.add_argument(
        "-t",
        "--StackTypes",
        help="Limit generation to EnvRoles of the specified StackTypes",
        nargs="+",
        default=[],
    )

    args = parser.parse_args()

    try:
        cfg.debug = args.Debug
    except Exception:
        cfg.debug = None
    cfg.regions = args.Regions

    return args



import sys
import os
import mmap

from . import cfg


def get_brands():
    brand_int = os.listdir(cfg.PATH_INT)
    brand_ext = os.listdir(cfg.PATH_EXT)

    brands = set(brand_int + brand_ext)

    brands.remove("BASE")

    return brands


def build_discover_map(brand, stacktypes, envroles):
    if not stacktypes and not envroles:
        # include all roles in all stacktypes
        stacktypes = cfg.STACK_TYPES

    roles = []

    path_int = os.path.join(cfg.PATH_INT, brand)
    path_ext = os.path.join(cfg.PATH_EXT, brand)

    for n in [path_int, path_ext]:
        for root, directories, filenames in os.walk(n, topdown=True):
            try:
                directories.remove("UNUSED")
            except Exception:
                pass
            for filename in filenames:
                root_dir = os.path.basename(root)
                stacktype = root_dir.lower()
                role = os.path.splitext(os.path.basename(filename))[0]
                if (
                    root_dir.isupper()
                    and not filename.startswith(".")
                    and not filename == "TYPE.yml"
                    and (stacktype in stacktypes or role in envroles)
                ):
                    roles.append((stacktype, role))

    if brand == "BASE":
        for n in ext_brands:
            add_to_map(n, roles)
    else:
        add_to_map(brand, roles)


def add_to_map(brand, roles):
    global discover_map

    try:
        discover_map[brand].extend(roles)
    except Exception:
        discover_map[brand] = roles


def discover(brands, envroles, stacktypes):
    global discover_map
    global ext_brands

    discover_map = {}

    if brands:
        ext_brands = list(brands)
    else:
        ext_brands = get_brands()
        brands = list(ext_brands)

    brands.append("BASE")

    for brand in brands:
        build_discover_map(brand, stacktypes, envroles)

    return discover_map



import os
import sys
from functools import reduce

cwd = os.getcwd()
sys.path.append(os.path.join(cwd, "lib"))

no_override = False

Parameters = {}
Conditions = {}
Mappings = {}
Resources = {}
Outputs = {}

OBJS = {}

PATH_INT = os.path.join(os.path.dirname(os.path.realpath(__file__)), "cfg")
PATH_INT = os.path.normpath(PATH_INT)

PATH_EXT = os.path.join(cwd, "cfg")
PATH_EXT = os.path.normpath(PATH_EXT)

STACK_TYPES = [
    "agw",
    "alb",
    "cch",
    "clf",
    "ec2",
    "ecr",
    "ecs",
    "rds",
    "res",
    "tsk",
    "lbd",
]

MAX_SECURITY_GROUPS = 4
# SECURITY_GROUPS_DEFAULT equals list of empty values
# (Ex. for "MAX_SECURITY_GROUPS = 3" we have "SECURITY_GROUPS_DEFAULT = ',,'")
SECURITY_GROUPS_DEFAULT = reduce(
    lambda a, i: f",{a}", list(range(MAX_SECURITY_GROUPS - 1)), ""
)

ENV_BASE = ["dev", "stg", "prd"]

DEFAULT_REGIONS = [
    "eu-west-1",
    "us-east-1",
]

AZones = {
    "MAX": 6,
    "default": 3,
    "us-east-1": 6,
    "us-west-1": 2,
    "us-west-2": 4,
    "ca-central-1": 2,
    "sa-east-1": 2,
    "cn-north-1": 2,
    "ap-northeast-3": 1,
}

AZoneNames = ["A", "B", "C", "D", "E", "F"]

VPC_DEFAULT_CIDR_BLOCK = "10.80.0.0/16"
VPC_DEFAULT_CIDR_BLOCK_PREFIX = ".".join(VPC_DEFAULT_CIDR_BLOCK.split(".")[0:2])
VPC_DEFAULT_SUBNETS_CIDR_BLOCK_PRIVATE = [
    f"{VPC_DEFAULT_CIDR_BLOCK_PREFIX}.{i * 16}.0/20" for i in range(AZones["MAX"])
]
VPC_DEFAULT_SUBNETS_CIDR_BLOCK_PUBLIC = [
    f"{VPC_DEFAULT_CIDR_BLOCK_PREFIX}.{i + 200}.0/24" for i in range(AZones["MAX"])
]

TROPO_CLASS_TO_CFM = {
    "AWS::WAFv2::WebACL": {
        "WebACLRule": "Rule",
    }
}

EVAL_TROPO_FUNCTIONS_IN_CFG = (
    "Sub(",
    "Ref(",
    "GetAtt(",
    "Split(",
    "Select(",
    "Export(",
    "ImportValue(",
    "Join(",
    "Base64(",
    "If(",
    "Equals(",
    "Not(",
    "GetAZs(",
    "Tags(",
)

EVAL_PYTHON_FUNCTIONS_IN_CFG = (
    "cfg.",
    "get_expvalue(",
    "get_subvalue(",
    "get_endvalue(",
    "get_resvalue(",
    "get_condition(",
    "dict(",
    "eval(",
    "str(",
    "int(",
    "list(",
    "getattr(",
    "SG_SecurityGroups",
)

EVAL_FUNCTIONS_IN_CFG = EVAL_PYTHON_FUNCTIONS_IN_CFG + EVAL_TROPO_FUNCTIONS_IN_CFG

CLF_PATH_PATTERN_REPLACEMENT = {
    "/": "SLASH",
    "*": "STAR",
    "-": "HYPH",
    "?": "QUEST",
    ".": "DOT",
    "_": "USCORE",
}

MERGE_RP_KEEP_AS_LIST = ("IBOX_LINKED_OBJ",)

INSTANCE_SIZES = [
    "nano",
    "micro",
    "small",
    "medium",
    "large",
    "xlarge",
    "2xlarge",
    "4xlarge",
    "8xlarge",
    "12xlarge",
    "16xlarge",
    "24xlarge",
    "32xlarge",
    "48xlarge",
]

INSTANCE_FAMILY = [
    {
        "Name": "t2",
        "Min": "nano",
        "Max": "2xlarge",
    },
    {
        "Name": "m4",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c4",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "r4",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "t3",
        "Min": "nano",
        "Max": "2xlarge",
    },
    {
        "Name": "m5",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c5",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "r5",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "t3a",
        "Min": "nano",
        "Max": "2xlarge",
    },
    {
        "Name": "t4g",
        "Min": "nano",
        "Max": "2xlarge",
    },
    {
        "Name": "m5a",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c5a",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "r5a",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "m6i",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c6i",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "r6i",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "m6a",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c6a",
        "Min": "large",
        "Max": "4xlarge",
    },
    {
        "Name": "c7g",
        "Min": "large",
        "Max": "4xlarge",
    },
]

# override previous cfg with an External one
try:
    with open(os.path.join(cwd, "lib", "cfgExt.py")) as f:
        exec(f.read())
except FileNotFoundError:
    pass

# build instances list
def build_instance_list():
    family_instances_list = []
    for family in INSTANCE_FAMILY:
        name = family["Name"]
        min_size = INSTANCE_SIZES.index(family["Min"])
        max_size = INSTANCE_SIZES.index(family["Max"])

        for s in INSTANCE_SIZES[min_size : max_size + 1]:
            family_instances_list.append(f"{name}.{s}")

    return family_instances_list


INSTANCE_LIST = ["default"] + build_instance_list()

# Order is VERY important do not CHANGE it!
CFG_TO_FUNC = {
    "MappingClass": {"module": "mappings", "func": "Mappings"},
    "Parameter": {"module": "cloudformation", "func": "CFM_Parameters"},
    "Condition": {"module": "cloudformation", "func": "CFM_Conditions"},
    "Mapping": {"module": "cloudformation", "func": "CFM_Mappings"},
    "ApiGatewayAccount": {"module": "joker", "func": ("apigateway", "Account")},
    "ApiGatewayApiKey": {"module": "joker", "func": ("apigateway", "ApiKey")},
    "ApiGatewayBasePathMapping": {
        "module": "joker",
        "func": ("apigateway", "BasePathMapping"),
    },
    "ApiGatewayDeployment": {
        "module": "joker",
        "func": ("apigateway", "Deployment"),
        "dep": ["ApiGatewayStage"],
    },
    "ApiGatewayDomainName": {"module": "joker", "func": ("apigateway", "DomainName")},
    "ApiGatewayMethod": {"module": "joker", "func": ("apigateway", "Method")},
    "ApiGatewayRestApi": {"module": "joker", "func": ("apigateway", "RestApi")},
    "ApiGatewayResource": {"module": "joker", "func": ("apigateway", "Resource")},
    "ApiGatewayStage": {"module": "joker", "func": ("apigateway", "Stage")},
    "ApiGatewayUsagePlan": {"module": "joker", "func": ("apigateway", "UsagePlan")},
    "ApiGatewayUsagePlanKey": {
        "module": "joker",
        "func": ("apigateway", "UsagePlanKey"),
    },
    "ApplicationAutoScalingScalableTarget": {
        "module": "joker",
        "func": ("applicationautoscaling", "ScalableTarget"),
    },
    "ApplicationAutoScalingScalingPolicy": {
        "module": "joker",
        "func": ("applicationautoscaling", "ScalingPolicy"),
    },
    "AutoScalingLifecycleHook": {
        "module": "joker",
        "func": ("autoscaling", "LifecycleHook"),
    },
    "AutoScalingGroup": {
        "module": "autoscaling",
        "func": "AS_Autoscaling",
        "dep": ["SecurityGroups"],
    },
    "AutoScalingScalingPolicy": {
        "module": "joker",
        "func": ("autoscaling", "ScalingPolicy"),
    },
    "AutoScalingScheduledAction": {
        "module": "joker",
        "func": ("autoscaling", "ScheduledAction"),
    },
    "CertificateManagerCertificate": {
        "module": "joker",
        "func": ("certificatemanager", "Certificate"),
    },
    "CloudFrontCachePolicy": {"module": "joker", "func": ("cloudfront", "CachePolicy")},
    "CloudFrontDistribution": {
        "module": "cloudfront",
        "func": "CF_CloudFront",
        "dep": [
            "ElasticLoadBalancingV2Listener",
            "LoadBalancer",
            "CloudFrontLambdaFunctionAssociation",
        ],
    },
    "CloudFrontFunction": {
        "module": "joker",
        "func": ("cloudfront", "Function"),
    },
    "CloudFrontLambdaFunctionAssociation": {
        "module": "joker",
        "func": ("cloudfront", "LambdaFunctionAssociation"),
    },
    "CloudFrontOriginAccessIdentity": {
        "module": "joker",
        "func": ("cloudfront", "CloudFrontOriginAccessIdentity"),
        "dep": ["S3Bucket"],
    },
    "CloudFrontOriginRequestPolicy": {
        "module": "joker",
        "func": ("cloudfront", "OriginRequestPolicy"),
    },
    "CloudFrontResponseHeadersPolicy": {
        "module": "joker",
        "func": ("cloudfront", "ResponseHeadersPolicy"),
    },
    "CloudWatchAlarm": {
        "module": "joker",
        "func": ("cloudwatch", "Alarm"),
        "dep": ["LoadBalancer", "SQSQueue"],
    },
    "CodeDeployApplication": {"module": "joker", "func": ("codedeploy", "Application")},
    "CodeDeployDeploymentGroup": {
        "module": "joker",
        "func": ("codedeploy", "DeploymentGroup"),
    },
    "DynamoDBTable": {"module": "joker", "func": ("dynamodb", "Table")},
    "EC2EIP": {"module": "joker", "func": ("ec2", "EIP")},
    "EC2InternetGateway": {"module": "joker", "func": ("ec2", "InternetGateway")},
    "EC2NatGateway": {"module": "joker", "func": ("ec2", "NatGateway")},
    "EC2Route": {"module": "joker", "func": ("ec2", "Route")},
    "EC2RouteTable": {"module": "joker", "func": ("ec2", "RouteTable")},
    "EC2SecurityGroup": {
        "module": "ec2",
        "func": "SG_SecurityGroup",
        "dep": ["ECSTaskDefinition"],
    },
    "EC2SecurityGroupIngress": {
        "module": "ec2",
        "func": "SG_SecurityGroupIngresses",
        "dep": ["ElasticLoadBalancingV2Listener"],
    },
    "EC2Subnet": {"module": "ec2", "func": "EC2_Subnet"},
    "EC2VPC": {"module": "joker", "func": ("ec2", "VPC")},
    "EC2VPCEndpoint": {"module": "joker", "func": ("ec2", "VPCEndpoint")},
    "EC2VPCGatewayAttachment": {
        "module": "joker",
        "func": ("ec2", "VPCGatewayAttachment"),
    },
    "EC2VPCPeeringConnection": {
        "module": "joker",
        "func": ("ec2", "VPCPeeringConnection"),
    },
    "ECRRepository": {"module": "ecr", "func": "ECR_Repositories"},
    "ECSCapacityProvider": {"module": "joker", "func": ("ecs", "CapacityProvider")},
    "ECSCluster": {"module": "joker", "func": ("ecs", "Cluster")},
    "ECSClusterCapacityProviderAssociations": {
        "module": "joker",
        "func": ("ecs", "ClusterCapacityProviderAssociations"),
    },
    "ECSService": {
        "module": "joker",
        "func": ("ecs", "Service"),
        "dep": ["SecurityGroups"],
    },
    "ECSTaskDefinition": {"module": "ecs", "func": "ECS_TaskDefinition"},
    "EFSAccessPoint": {"module": "joker", "func": ("efs", "AccessPoint")},
    "EFSFileSystem": {"module": "efs", "func": "EFS_FileStorage"},
    "ElastiCacheCacheCluster": {
        "module": "joker",
        "func": ("elasticache", "CacheCluster"),
    },
    "ElastiCacheReplicationGroup": {
        "module": "joker",
        "func": ("elasticache", "ReplicationGroup"),
    },
    "ElastiCacheSubnetGroup": {
        "module": "joker",
        "func": ("elasticache", "SubnetGroup"),
    },
    "ElasticLoadBalancingV2LoadBalancer": {
        "module": "joker",
        "func": ("elasticloadbalancingv2", "LoadBalancer"),
    },
    "ElasticLoadBalancingV2Listener": {
        "module": "joker",
        "func": ("elasticloadbalancingv2", "Listener"),
        "dep": ["LoadBalancer"],
    },
    "ElasticLoadBalancingV2ListenerRule": {
        "module": "joker",
        "func": ("elasticloadbalancingv2", "ListenerRule"),
    },
    "ElasticLoadBalancingV2TargetGroup": {
        "module": "joker",
        "func": ("elasticloadbalancingv2", "TargetGroup"),
        # need this for override on ContainerDefinitions1ContainerPort
        "dep": ["ECSTaskDefinition"],
    },
    "EventsRule": {
        "module": "joker",
        "func": ("events", "Rule"),
        "dep": ["SecurityGroups"],
    },
    "IAMGroup": {"module": "iam", "func": "IAM_Groups"},
    "IAMInstanceProfile": {"module": "joker", "func": ("iam", "InstanceProfile")},
    "IAMManagedPolicy": {"module": "joker", "func": ("iam", "ManagedPolicy")},
    "IAMPolicy": {
        "module": "joker",
        "func": ("iam", "PolicyType"),
        "dep": ["S3Bucket"],
    },
    "IAMRole": {
        "module": "joker",
        "func": ("iam", "Role"),
        "dep": ["Lambda", "S3Bucket"],
    },
    "IAMUser": {"module": "iam", "func": "IAM_Users"},
    "IAMUserToGroupAddition": {"module": "iam", "func": "IAM_UserToGroupAdditions"},
    "KMSAlias": {"module": "joker", "func": ("kms", "Alias")},
    "KMSKey": {"module": "joker", "func": ("kms", "Key")},
    "Lambda": {"module": "joker", "func": ("awslambda", "Function")},
    "LambdaEventInvokeConfig": {
        "module": "joker",
        "func": ("awslambda", "EventInvokeConfig"),
    },
    "LambdaEventSourceMapping": {
        "module": "joker",
        "func": ("awslambda", "EventSourceMapping"),
    },
    "LambdaLayerVersion": {"module": "joker", "func": ("awslambda", "LayerVersion")},
    "LambdaLayerVersionPermission": {
        "module": "joker",
        "func": ("awslambda", "LayerVersionPermission"),
        "dep": ["LambdaLayerVersion"],
    },
    "LambdaPermission": {
        "module": "joker",
        "func": ("awslambda", "Permission"),
        "dep": ["SNSSubscription", "Lambda", "EventsRule", "SchedulerSchedule"],
    },
    "LambdaVersion": {
        "module": "joker",
        "func": ("awslambda", "Version"),
        "dep": ["Lambda"],
    },
    "LoadBalancer": {
        "module": "elasticloadbalancing",
        "func": "LB_ElasticLoadBalancing",
    },
    "LogsLogGroup": {
        "module": "joker",
        "func": ("logs", "LogGroup"),
        "dep": ["Lambda"],
    },
    "RDSDBInstance": {"module": "rds", "func": "RDS_DB"},
    "RDSDBSubnetGroup": {"module": "joker", "func": ("rds", "DBSubnetGroup")},
    "Route53HostedZone": {"module": "joker", "func": ("route53", "HostedZone")},
    "Route53RecordSet": {
        "module": "joker",
        "func": ("route53", "RecordSetType"),
        "dep": [
            "ApiGatewayDomainName",
            "DBInstance",
            "EFSFileSystem",
            "LoadBalancer",
            "CloudFrontDistribution",
        ],
    },
    "Route53ResolverEndpoint": {
        "module": "joker",
        "func": ("route53resolver", "ResolverEndpoint"),
    },
    "Route53ResolverRule": {
        "module": "joker",
        "func": ("route53resolver", "ResolverRule"),
    },
    "Route53ResolverRuleAssociation": {
        "module": "joker",
        "func": ("route53resolver", "ResolverRuleAssociation"),
    },
    "S3Bucket": {"module": "s3", "func": "S3_Buckets"},
    "SecurityGroups": {
        "module": "ec2",
        "func": "SG_SecurityGroups",
        "dep": ["LoadBalancer"],
    },
    "ServiceDiscoveryPublicDnsNamespace": {
        "module": "joker",
        "func": ("servicediscovery", "PublicDnsNamespace"),
    },
    "ServiceDiscoveryService": {
        "module": "joker",
        "func": ("servicediscovery", "Service"),
    },
    "SchedulerSchedule": {
        "module": "joker",
        "func": ("scheduler", "Schedule"),
        "dep": ["SecurityGroups"],
    },
    "SQSQueue": {"module": "joker", "func": ("sqs", "Queue")},
    "SQSQueuePolicy": {
        "module": "joker",
        "func": ("sqs", "QueuePolicy"),
        "dep": ["SNSSubscription"],
    },
    "SNSSubscription": {"module": "joker", "func": ("sns", "SubscriptionResource")},
    "SNSTopic": {"module": "joker", "func": ("sns", "Topic")},
    "SSMParameter": {"module": "joker", "func": ("ssm", "Parameter")},
    "WafByteMatchSet": {
        "module": "waf",
        "func": ["WAF_GlobalByteMatchSets", "WAF_RegionalByteMatchSets"],
    },
    "WafIPSet": {"module": "waf", "func": ["WAF_GlobalIPSets", "WAF_RegionalIPSets"]},
    "WafRule": {"module": "waf", "func": ["WAF_GlobalRules", "WAF_RegionalRules"]},
    "WafWebAcl": {
        "module": "waf",
        "func": ["WAF_GlobalWebAcls", "WAF_RegionalWebAcls"],
    },
    "WAFv2IPSet": {
        "module": "joker",
        "func": ("wafv2", "IPSet"),
    },
    "WAFv2WebACL": {
        "module": "joker",
        "func": ("wafv2", "WebACL"),
    },
    # CloudformationCustomResource begin here
    "CCRFargateSpot": {
        "module": "cloudformation",
        "func": "CFM_CustomResourceFargateSpot",
    },
    "CCRLightHouse": {
        "module": "cloudformation",
        "func": "CFM_CustomResourceLightHouse",
    },
    # ReplicateRegions need to be the last one
    "CCRReplicateRegions": {
        "module": "cloudformation",
        "func": "CFM_CustomResourceReplicator",
    },
    # Output need to be last line
    "Output": {"module": "cloudformation", "func": "CFM_Outputs"},
}
